#  Copyright (c) 2022 by Amplo.

from copy import deepcopy
from warnings import warn

import numpy as np
from sklearn.metrics import (
    accuracy_score,
    confusion_matrix,
    f1_score,
    log_loss,
    max_error,
    mean_absolute_error,
    mean_absolute_percentage_error,
    mean_squared_error,
    precision_score,
    r2_score,
    roc_auc_score,
)
from sklearn.model_selection import KFold, StratifiedKFold

from amplo.base import LoggingMixin

__all__ = ["ModelValidator"]


class ModelValidator(LoggingMixin):
    """
    Model validator.

    Parameters
    ----------
    cv_splits : int, default: 5
        Number of cross validation splits.
    cv_shuffle : bool, default: True
        Whether to shuffle the data for cross validation.
    verbose : {0, 1, 2}, default: 1
        Verbosity for logging.
    """

    def __init__(self, cv_splits=5, cv_shuffle=True, verbose=1):
        super().__init__(verbose=verbose)
        self.cv_splits = cv_splits
        self.cv_shuffle = cv_shuffle

    def validate(self, model, x, y, mode):
        """
        Validate model and return performance metrics.

        Parameters
        ----------
        model :
            Model to be validated.
        x : array_like
            Training data for validation.
        y : array_like
            Target data for validation.
        mode : {"classification", "regression"}
            Model mode.

        Returns
        -------
        metrics : dict
            Cross validated model performance metrics.
        """
        x = np.asarray(x)
        y = np.asarray(y)

        # Calculate metrics
        if mode == "classification":
            metrics = self._validate_classification(model, x, y)
        elif mode == "regression":
            metrics = self._validate_regression(model, x, y)
        else:
            warn("Invalid mode for validation.", UserWarning)
            metrics = {}

        # Logging
        for name, values in metrics.items():
            # We expect a mean and std value
            if not isinstance(values, list) or np.array(values).size != 2:
                continue

            name = f"{name.replace('_', ' ').title()}:".ljust(20)
            self.logger.info(f"{name} {values[0]:.2f} \u00B1 {values[1]:.2f}")

        cm = np.array(metrics.get("confusion_matrix", []))
        if mode == "classification" and cm.shape == (2, 2, 2):
            cm_text = []
            for mean, std in zip(cm[0].ravel(), cm[1].ravel()):
                cm_text.append(f"{mean:.2f} \u00B1 {std:.2f}".ljust(12))

            for line in [
                "Confusion Matrix:",
                "  actual / predicted  |    Negative    |    Positive    |",
                f"            Negative  |  {cm_text[0]}  |  {cm_text[1]}  |",
                f"            Positive  |  {cm_text[2]}  |  {cm_text[3]}  |",
            ]:
                self.logger.info(line)

        return metrics

    def _validate_classification(self, model, x, y):
        """
        Cross validate classification (binary or multiclass) model.

        Parameters
        ----------
        model :
            Model to be validated.
        x : np.ndarray
            Training data for validation.
        y : np.ndarray
            Target data for validation.

        Returns
        -------
        metrics : dict
            Cross validated performance metrics of model.
        """

        labels = np.unique(y)
        is_binary = labels.size == 2

        # Initialize metrics
        accuracy = []
        precision = []
        f1 = []
        cm = []
        log_loss_ = []

        # Binary metrics
        roc_auc = []
        sensitivity = []
        specificity = []

        # Modelling
        cv = StratifiedKFold(
            n_splits=self.cv_splits, shuffle=self.cv_shuffle, random_state=38234
        )
        for t, v in cv.split(x, y):
            xt, yt = x[t], y[t].reshape(-1)
            xv, yv = x[v], y[v].reshape(-1)
            cv_model = deepcopy(model)
            cv_model.fit(xt, yt)
            yv_pred = cv_model.predict(xv).reshape(-1)

            # Metrics
            accuracy.append(accuracy_score(yv, yv_pred))
            cm.append(confusion_matrix(yv, yv_pred, labels=labels))
            if hasattr(cv_model, "predict_proba"):
                log_loss_.append(log_loss(yv, cv_model.predict_proba(xv)))

            if is_binary:
                roc_auc.append(roc_auc_score(yv, yv_pred, labels=labels))
                tn, fp, fn, tp = cm[-1].ravel()
                if tp + fp > 0:
                    precision.append(precision_score(yv, yv_pred, labels=labels))
                if tp + fn > 0:
                    sensitivity.append(tp / (tp + fn))
                if tn + fp > 0:
                    specificity.append(tn / (tn + fp))
                if tp + fp > 0 and tp + fn > 0:
                    f1.append(f1_score(yv, yv_pred, labels=labels))
            else:
                precision.append(
                    precision_score(yv, yv_pred, labels=labels, average="weighted")
                )
                f1.append(f1_score(yv, yv_pred, labels=labels, average="weighted"))

        # Summarize metrics
        metrics = {}
        for name, values in [
            ("accuracy", accuracy),
            ("precision", precision),
            ("sensitivity", sensitivity),
            ("specificity", specificity),
            ("f1_score", f1),
            ("log_loss", log_loss_),
        ]:
            if len(values) == 0:
                # Skip empty values (for multi-class)
                continue
            metrics[name] = [np.mean(values), np.std(values)]

        # Add confusion matrix metrics
        cm_totals = np.sum(cm, axis=(1, 2), keepdims=True)
        cm_means = np.mean(cm / cm_totals, axis=0)
        cm_stds = np.std(cm / cm_totals, axis=0)
        del cm_totals
        metrics["confusion_matrix"] = [cm_means.tolist(), cm_stds.tolist()]

        if is_binary:
            metrics["true_positives"] = [cm_means[0, 0], cm_stds[0, 0]]
            metrics["false_positives"] = [cm_means[0, 1], cm_stds[0, 1]]
            metrics["true_negatives"] = [cm_means[1, 0], cm_stds[1, 0]]
            metrics["false_negatives"] = [cm_means[1, 1], cm_stds[1, 1]]

        return metrics

    def _validate_regression(self, model, x, y):
        """
        Cross validate regression model.

        Parameters
        ----------
        model :
            Model to be validated.
        x : np.ndarray
            Training data for validation.
        y : np.ndarray
            Target data for validation.

        Returns
        -------
        metrics : dict
            Cross validated performance metrics of model.
        """

        # Initialize metrics
        mae = []
        mse = []
        r2 = []
        max_error_ = []
        mean_rel_error = []

        # Modelling
        cv = KFold(n_splits=self.cv_splits, shuffle=self.cv_shuffle, random_state=29283)
        for t, v in cv.split(x, y):
            xt, yt = x[t], y[t].reshape(-1)
            xv, yv = x[v], y[v].reshape(-1)
            cv_model = deepcopy(model)
            cv_model.fit(xt, yt)
            yv_pred = cv_model.predict(xv).reshape(-1)

            # Metrics
            mae.append(mean_absolute_error(yv, yv_pred))
            mse.append(mean_squared_error(yv, yv_pred))
            r2.append(r2_score(yv, yv_pred))
            max_error_.append(max_error(yv, yv_pred))
            mean_rel_error.append(mean_absolute_percentage_error(yv, yv_pred))

        # Summarize metrics
        metrics = {}
        for name, values in [
            ("mean_absolute_error", mae),
            ("mean_squared_error", mse),
            ("r2_score", r2),
            ("max_error", max_error_),
            ("mean_relative_error", mean_rel_error),
        ]:
            metrics[name] = [np.mean(values), np.std(values)]

        return metrics



#  Copyright (c) 2022 by Amplo.

import pandas as pd


class Standardizer:
    def __init__(self, float_cols: list = None):
        self.float_cols = float_cols
        self.means = None
        self.stds = None

    def fit_transform(self, df: pd.DataFrame) -> pd.DataFrame:
        # Fit Input
        self.fit(df)
        df = self.transform(df)
        return df

    def fit(self, df: pd.DataFrame):
        self.means = df[self.float_cols].mean(axis=0)
        self.stds = df[self.float_cols].std(axis=0)
        self.stds[abs(self.stds) <= 1e-9] = 1

    def transform(self, df: pd.DataFrame):
        df[self.float_cols] = (df[self.float_cols] - self.means) / self.stds
        return df



#  Copyright (c) 2022 by Amplo.

from __future__ import annotations

import json
import os
import re
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, TextIO
from warnings import warn

import requests

from amplo.api._base import BaseRequestAPI
from amplo.utils.util import check_dtypes, deprecated

if TYPE_CHECKING:
    from requests import Response

__all__ = [
    "AmploPlatformAPI",
    "upload_model",
    "report_training_fail",
    "PlatformSynchronizer",
]


_PLATFORM_HOST = "https://platform.amplo.ch"
_PLATFORM_TOKEN_OS = "AMPLO_PLATFORM_STRING"


def _format_version(version: int | str | None) -> str | None:
    """
    If provided, validates and formats a model version.

    Parameters
    ----------
    version : int or str
        Version to be validated

    Returns
    -------
    str
        If version is not None, returns formatted model version (e.g. "v1", "v2", ...).

    Raises
    ------
    ValueError
        When `version` is malformed, i.e. not mutable to an integer.
    """

    if version is None:
        return None

    # Stringify and append "v" in front of it
    version = re.sub("^v+", "v", f"v{version}")

    # Validity check
    try:
        _ = int(version.removeprefix("v"))
    except ValueError as err:
        raise ValueError(f"Parameter `version` is malformed: {err}")

    return version


class AmploPlatformAPI(BaseRequestAPI):
    """
    Helper class for working with Amplo's Platform API.

    Parameters
    ----------
    host : str
        Amplo platform host.
    access_token : str
        Access token, a.k.a. "X-Api-Key", for platform.
    """

    def __init__(self, host: str, access_token: str):
        super().__init__(host, access_token)

    def _authorization_header(self) -> dict:
        return {"X-Api-Key": self.access_token}

    @classmethod
    def from_os_env(
        cls, host: str | None = None, access_token_os: str | None = None
    ) -> AmploPlatformAPI:
        """
        Instantiate the class using os environment strings.

        Parameters
        ----------
        host : str, default: _PLATFORM_HOST
            Amplo host. Not from os environment!
        access_token_os : str, default: _PLATFORM_TOKEN_OS
            Key in the os environment for the platform access token.

        Returns
        -------
        AmploPlatformAPI

        Raises
        ------
        KeyError
            When a os variable is not set.
        """

        access_token_os = access_token_os or _PLATFORM_TOKEN_OS

        host = host or _PLATFORM_HOST
        access_token = os.environ[access_token_os]
        return cls(host, access_token)

    def list_models(
        self,
        team: str | None = None,
        machine: str | None = None,
        service: str | None = None,
        issue: str | None = None,
        **more_params,
    ) -> list[dict]:
        params = {
            "team": team,
            "machine": machine,
            "category": service,
            "name": issue,
            **more_params,
        }
        return self.request("get", "models", params=params).json()

    def list_trainings(
        self,
        team: str | None = None,
        machine: str | None = None,
        service: str | None = None,
        issue: str | None = None,
        version: str | int | None = None,
        **more_params,
    ) -> list[dict]:
        version = _format_version(version)
        params = {
            "team": team,
            "machine": machine,
            "category": service,
            "model": issue,
            "version": version,
            **more_params,
        }
        return self.request("get", "trainings", params=params).json()

    def upload_training(
        self,
        team: str,
        machine: str,
        training_id: int,
        files: list[TextIO] | None = None,
        status: int | None = None,
    ) -> Response:
        data = {
            "team": team,
            "machine": machine,
            "id": training_id,
            "files": files,
            "new_status": status,
        }
        return super().request("put", "trainings", data=data)

    def get_datalogs(
        self,
        team: str,
        machine: str,
        category: str | None = None,
        **more_params,
    ) -> list[dict]:
        if more_params.get("filename", False):
            warn(
                "Found 'filename' key in 'more_params'. "
                "Consider using the method 'get_datalog' instead."
            )
        params = {"team": team, "machine": machine, "category": category, **more_params}
        return self.request("get", "datalogs", params=params).json()

    def get_datalog(
        self,
        team: str,
        machine: str,
        filename: str,
    ) -> dict:
        params = {"team": team, "machine": machine, "filename": filename}
        return self.request("get", "datalogs", params=params).json()


def upload_model(
    model_id: int,
    model_dir: str | Path,
    team: str,
    machine: str,
    service: str,
    issue: str,
    version: str | int,
    *,
    host: str | None = None,
    access_token_os: str | None = None,
) -> Response:
    """
    Uploads a trained model to the Amplo platform.

    Notes
    -----
    Make sure to have set the following environment variables:
        - ``AMPLO_PLATFORM_STRING`` (access token for platform).

    Parameters
    ----------
    model_id : int
        Model training ID.
    model_dir : str or Path
        Model directory which contains a "Production/v{version}/Settings.json" and
        "Production/v{version}/Model.joblib" file.
    team : str
        Name of the team.
    machine : str
        Name of the machine.
    service : str
        Name of the service (a.k.a. category).
    issue : str
        Name of the issue (a.k.a. model).
    version : str or int
        Model version ID, e.g. "v1".
    host : str, default: _PLATFORM_HOST
        Amplo platform host. Not from os environment!
    access_token_os : str, default: _PLATFORM_TOKEN_OS
        Key in the os environment for the platform access token.

    Returns
    -------
    requests.Response

    Raises
    ------
    ValueError
        When no training exists on the platform with the given parameters.
    """

    check_dtypes(
        ("model_id", model_id, int),
        ("model_dir", model_dir, (str, Path)),
        ("team", team, str),
        ("machine", machine, str),
        ("service", service, str),
        ("issue", issue, str),
        ("version", version, (str, int)),
    )

    # Set model directory to the requested version
    model_dir = Path(model_dir) / "Production" / str(_format_version(version))

    # Check directory
    model_dir = Path(model_dir)
    if not model_dir.is_dir():
        raise NotADirectoryError(f"Invalid `model_dir` directory: {model_dir}")

    # Check and set model files
    model_files = []
    for file in ("Settings.json", "Model.joblib"):
        if not (model_dir / file).exists():
            raise FileNotFoundError(f"File '{file}' not found in '{model_dir}'.")
        io = open(model_dir / file, "r")
        model_files.append(io)

    # Check that the training of the model exists
    api = AmploPlatformAPI.from_os_env(host, access_token_os)
    trainings = api.list_trainings(team, machine, service, issue, version, id=model_id)
    if len(trainings) != 1:
        raise ValueError("There exists no training with the given parameters.")

    return api.upload_training(team, machine, model_id, model_files)


def report_training_fail(
    team: str,
    machine: str,
    training_id: int,
    *,
    host: str | None = None,
    access_token_os: str | None = None,
):
    """
    Report training status "Failed" to the platform.

    Parameters
    ----------
    team : str
        Name of the team.
    machine : str
        Name of the machine.
    training_id : int, optional
        Model training ID.
    host : str, default: _PLATFORM_HOST
        Amplo platform host. Not from os environment!
    access_token_os : str, default: _PLATFORM_TOKEN_OS
        Key in the os environment for the platform access token.
    """

    api = AmploPlatformAPI.from_os_env(host, access_token_os)
    api.upload_training(team, machine, training_id, status=4)  # 4 == "Failed"

    return None


@deprecated("Use `upload_model` (or directly `AmploPlatformAPI`) instead.")
class PlatformSynchronizer:
    def __init__(
        self,
        connection_string_name="AMPLO_PLATFORM_STRING",
        api_url="https://platform.amplo.ch/api",
        verbose=0,
    ):
        """
        Connector to Amplo`s platform for uploading trained models

        Parameters
        ----------
        connection_string_name : str
        api_url : str
        verbose : int
        """
        self.api_url = re.sub(r"/$", "", api_url)
        self.api_key = os.getenv(connection_string_name)
        self.verbose = int(verbose)

    def upload_model(self, issue_dir, team, machine, service, issue, version, model_id):
        """
        Upload trained model

        Parameters
        ----------
        issue_dir : str or Path
            Model issue directory
        team : str
            Team name
        machine : str
            Machine name
        service : str
            Service name
        issue : str
            Issue name
        version : int
            Version of production model to upload
        model_id : int
            Model identifier for platform.

        Notes
        -----
        Data structures
            - Locally: ``Auto_ML / Production / vX / ...``
            - Cloud: ``Team / Machine / models / Diagnostics / vX / ...``
        """

        # Check input args
        if not os.path.isdir(issue_dir):
            raise FileNotFoundError("Model directory does not exist.")
        check_dtypes(
            ("team", team, str),
            ("machine", machine, str),
            ("service", service, str),
            ("issue", issue, str),
            ("version", version, int),
        )

        # Check file structure
        file_dir = Path(issue_dir) / "Production" / f"v{version}"
        if not os.path.isdir(file_dir):
            raise FileNotFoundError(
                f"Production files for version {version} don't exist."
            )
        files = [file_dir / file for file in ("Settings.json", "Model.joblib")]
        for file in files:
            if not file.exists():
                raise FileNotFoundError(f"File {file} doesn't exist in {file_dir}.")

        # Set info
        headers = {"X-Api-Key": self.api_key}
        data = {
            "team": team,  # team name
            "machine": machine,  # machine name
            "category": service,  # category
            "model": issue,
            "name": issue,  # model name (for PUT or POST, respectively)
            "version": f"v{version}",  # latest version
        }

        # Assert that model exists
        def model_exists(model_name: str):
            # GET request
            get_response = requests.get(
                self.api_url + "/models", params=data, headers=headers
            )
            if get_response.status_code != 200:
                raise requests.HTTPError(f"{get_response} {get_response.text}")

            # Get all existing models
            try:
                # Convert string to list of dict
                existing_models = json.loads(get_response.text)
                if isinstance(existing_models, str):
                    ValueError("Could not transform response.")
            except json.decoder.JSONDecodeError as err:
                warnings.warn(f"Could not decode GET response - {err}")
                # Set default value of ``existing_models`` to list of (empty) dict
                existing_models = list(dict())

            # Check existing models
            if not isinstance(existing_models, list):
                raise ValueError(f"Unexpected data type: {type(existing_models)}")
            for model in existing_models:
                if not isinstance(model, dict):
                    raise ValueError(f"Unexpected data type: {type(model)}")

            # Check if model exists
            return any(
                model.get("name", None) == model_name for model in existing_models
            )

        # POST when model yet doesn't exist
        if not model_exists(data["name"]):
            warnings.warn(
                f"Model {data['name']} did not yet exist. POST-ing a new model to the "
                f"platform..."
            )
            post_response = requests.post(
                self.api_url + "/models", data=data, headers=headers
            )
            # Assert that now model exists
            if not model_exists(data["name"]):
                raise requests.HTTPError(f"{post_response} {post_response.text}")

        # Add files and model id to data
        data.update({"files": [open(file, "r") for file in files], "id": model_id})
        # PUT a new version
        put_response = requests.put(
            self.api_url + "/trainings", data=data, headers=headers
        )
        if put_response.status_code != 200:
            raise requests.HTTPError(f"{put_response} {put_response.text}")

    def upload_latest_model(self, issue_dir, *args, **kwargs):
        """
        Same as `upload_model` but selects the latest version of production model

        Parameters
        ----------
        issue_dir : str or Path
            Model issue directory
        args
            Arguments to be passed to `upload_model`
        kwargs
            Keyword arguments to be passed to `upload model`
        """
        # Check input args
        assert os.path.isdir(issue_dir), "Model directory does not exist"

        # Check file structure
        all_version_dirs = [
            dir_
            for dir_ in (Path(issue_dir) / "Production").glob("v*")
            if dir_.is_dir()
        ]
        assert len(all_version_dirs) > 0, "No production data found"

        # Take the latest version and remove `v` in e.g. `v1`
        latest_version = int(sorted(all_version_dirs)[-1].name[1:])

        # Set version in kwargs and upload model
        kwargs.update({"version": latest_version})
        return self.upload_model(issue_dir, *args, **kwargs)



#  Copyright (c) 2022 by Amplo.

from __future__ import annotations

import copy
import itertools
import json
import os
import time
from datetime import datetime
from inspect import signature
from pathlib import Path
from typing import Any, Tuple, Union
from warnings import warn

import joblib
import numpy as np
import pandas as pd
from shap import TreeExplainer
from sklearn import metrics
from sklearn.model_selection import KFold, StratifiedKFold
from tqdm import tqdm

import amplo
from amplo import utils
from amplo.automl.data_exploring import DataExplorer
from amplo.automl.data_processing import DataProcessor
from amplo.automl.data_sampling import DataSampler
from amplo.automl.drift_detection import DriftDetector
from amplo.automl.feature_processing import FeatureProcessor
from amplo.automl.feature_processing.feature_processor import (
    get_required_columns,
    translate_features,
)
from amplo.automl.interval_analysis import IntervalAnalyser
from amplo.automl.modelling import Modeller
from amplo.automl.sequencing import Sequencer
from amplo.base.objects import LoggingMixin
from amplo.classification.stacking import StackingClassifier
from amplo.grid_search import ExhaustiveGridSearch, HalvingGridSearch, OptunaGridSearch
from amplo.observation import DataObserver, ModelObserver
from amplo.regression.stacking import StackingRegressor
from amplo.validation import ModelValidator

__all__ = ["Pipeline"]


class Pipeline(LoggingMixin):
    """
    Automated Machine Learning Pipeline for tabular data.

    The pipeline is designed for predictive maintenance application, failure
    identification, failure prediction, condition monitoring, and more.

    Parameters
    ----------
    # Main parameters
    main_dir : str, default: "Auto_ML/"
        Main directory of pipeline (for documentation).
    target : str, optional
        Column name of the output variable.
    name : str, default: "AutoML"
        Name of the project (for documentation).
    version : int, default: 1
        Pipeline version. Will automatically increment when a version exists.
    mode : {None, "classification", "regression"}, default: None
        Pipeline mode.
    objective : str, optional
        Objective for training.
        Default for classification: "neg_log_loss".
        Default for regression: "mean_square_error".
    verbose : int, default: 1
        Verbosity of logging.
    logging_to_file : bool, default: False
        Whether to write logging to a file
    logging_path : str, default: "AutoML.log"
        Write to logging to given path if ``logs_to_file`` is True.

    # Data processing
    int_cols : list of str, optional
        Column names of integer columns (if None will detect automatically).
    float_cols : list of str, optional
        Column names of float columns (if None will detect automatically).
    date_cols : list of str, optional
        Column names of datetime column (if None will detect automatically).
    cat_cols : list of str, optional
        Column names of categorical columns (if None will detect automatically).
    missing_values : {"remove", "interpolate", "mean", "zero"}, default: "zero"
        How to treat missing values.
    outlier_removal : {"clip", "boxplot", "z-score", "none"}, default: "clip"
        How to treat outliers.
    z_score_threshold : int, default: 4
        When ``outlier_removal`` is "z-score", the threshold is adaptable.
    include_output : bool, default: False
        Whether to include output in the training data (sensible only with sequencing).

    # Balancing
    balance : bool, default: False
        Whether to balance data.

    # Feature processing
    extract_features : bool, default: True
        Whether to use the FeatureProcessing module to extract features.
    information_threshold : float, default: 0.999
        Threshold for removing collinear features.
    feature_timeout : int, default: 3600
        Time budget for feature processing.

    # Interval analysis
    interval_analyse : bool, default: False
        Whether to use the IntervalAnalyser module.

    # Sequencing
    sequence : bool, default: False
        Whether to use the Sequencer module.
    seq_back : int or list of int, default: 1
        Input time indices.
        If int: includes that many samples backward.
        If list of int: includes all integers within the list.
    seq_forward : int or list of int, default: 1
        Output time indices.
        If int: includes that many samples forward.
        If list of int: includes all integers within the list.
    seq_shift : int, default: 0
        Shift input / output samples in time.
    seq_diff : {"none", "diff", "log_diff"}, default: "none"
        Difference the input and output.
    seq_flat : bool, default: True
        Whether to return a matrix (True) or a tensor (False).

    # Modelling
    standardize : bool, default: False
        Whether to standardize the input/output data.
    cv_shuffle : bool, default: True
        Whether to shuffle the samples during cross-validation.
    cv_splits : int, default: 10
        How many cross-validation splits to make.
    store_models : bool, default: False
        Whether to store all trained model files.

    # Grid search
    grid_search_type : {"exhaustive", "halving", "optuna"}, default: "optuna"
        Type of grid search to apply.
    grid_search_timeout : int, default: 3600
        Time budget for grid search (in seconds).
    n_grid_searches : int, default: 3
        Run grid search for the best `n_grid_searches` (model, feature set) pairs from
        initial modelling.
    n_trials_per_grid_search : int, default: 250
        Maximal number of trials/candidates for each grid search.

    # Stacking of models
    stacking : bool, default: False
        Whether to create a stacking model at the end.

    # Production
    preprocess_function : str, optional, default: None
        Add custom code for the prediction function - useful for production.
        Will be executed with `exec`, can be multi-line, uses data as input.

    # Flags
    plot_eda : bool, default: False
        Whether to run exploratory data analysis.
    process_data : bool, default: True
        Whether to force data processing.
    no_dirs : bool, default: False
        Whether to create files.

    # Other
    kwargs: Any
        Swallows all arguments that are not accepted. Warnings are raised if not empty.
    """

    def __init__(
        self,
        # Main settings
        main_dir: str = "Auto_ML/",
        target: str | None = None,
        name: str = "AutoML",
        version: int | None = None,
        mode: str | None = None,
        objective: str | None = None,
        verbose: int = 1,
        logging_to_file: bool = False,
        logging_path: str = "AutoML.log",
        *,
        # Data processing
        int_cols: list[str] | None = None,
        float_cols: list[str] | None = None,
        date_cols: list[str] | None = None,
        cat_cols: list[str] | None = None,
        missing_values: str = "zero",
        outlier_removal: str = "clip",
        z_score_threshold: int = 4,
        include_output: bool = False,
        # Balancing
        balance: bool = False,
        # Feature processing
        extract_features: bool = True,
        information_threshold: float = 0.999,
        feature_timeout: int = 3600,
        # Interval analysis
        interval_analyse: bool = True,
        # Sequencing
        sequence: bool = False,
        seq_back: int | list[int] = 1,
        seq_forward: int | list[int] = 1,
        seq_shift: int = 0,
        seq_diff: str = "none",
        seq_flat: bool = True,
        # Modelling
        standardize: bool = False,
        cv_shuffle: bool = True,
        cv_splits: int = 10,
        store_models: bool = False,
        # Grid search
        grid_search_type: str = "optuna",
        grid_search_timeout: int = 3600,
        n_grid_searches: int = 3,
        n_trials_per_grid_search: int = 250,
        # Stacking
        stacking: bool = False,
        # Production
        preprocess_function: str | None = None,
        # Flags
        plot_eda: bool = False,
        process_data: bool = True,
        no_dirs: bool = False,
        # Other
        **kwargs,
    ):
        # Get init parameters for `self.settings`
        sig, init_locals = signature(self.__init__), locals()
        init_params = {
            param.name: init_locals[param.name] for param in sig.parameters.values()
        }
        del sig, init_locals

        # Initialize Logger
        super().__init__(verbose=verbose)
        if logging_to_file:
            utils.logging.add_file_handler(logging_path)

        # Input checks: validity
        if mode not in (None, "regression", "classification"):
            raise ValueError("Supported models: {'regression', 'classification', None}")
        if not 0 < information_threshold < 1:
            raise ValueError("Information threshold must be within (0, 1) interval.")
        valid_gs_types = {"exhaustive", "halving", "optuna"}
        if (
            grid_search_type is not None
            and grid_search_type.lower() not in valid_gs_types
        ):
            raise ValueError(f"Grid search type must be one of: {valid_gs_types}")

        # Warn unused parameters
        if kwargs:
            warn(f"Got unexpected keyword arguments that are not handled: {kwargs}")

        # Input checks: advices
        if include_output and not sequence:
            warn("It is strongly advised NOT to include output without sequencing.")

        # Main settings
        self.main_dir = f"{Path(main_dir)}/"  # assert backslash afterwards
        self.target = target
        self.name = name
        self.version = version
        self.mode = mode
        self.objective = objective

        # Data processing
        self.int_cols = int_cols
        self.float_cols = float_cols
        self.date_cols = date_cols
        self.cat_cols = cat_cols
        self.missing_values = missing_values
        self.outlier_removal = outlier_removal
        self.z_score_threshold = z_score_threshold
        self.include_output = include_output

        # Balancing
        self.balance = balance

        # Feature processing
        self.extract_features = extract_features
        self.information_threshold = information_threshold
        self.feature_timeout = feature_timeout

        # Interval analysis
        self.use_interval_analyser = interval_analyse

        # Sequencing
        self.sequence = sequence
        self.sequence_back = seq_back
        self.sequence_forward = seq_forward
        self.sequence_shift = seq_shift
        self.sequence_diff = seq_diff
        self.sequence_flat = seq_flat

        # Modelling
        self.standardize = standardize
        self.cv_shuffle = cv_shuffle
        self.cv_splits = cv_splits
        self.store_models = store_models

        # Grid search
        self.grid_search_type = grid_search_type
        self.grid_search_timeout = grid_search_timeout
        self.n_grid_searches = n_grid_searches
        self.n_trials_per_grid_search = n_trials_per_grid_search

        # Stacking
        self.stacking = stacking

        # Production
        self.preprocess_function = preprocess_function

        # Flags
        self.plot_eda = plot_eda
        self.process_data = process_data
        self.no_dirs = no_dirs

        # Create directory
        if not self.no_dirs:
            self._create_dirs()
            self._load_version()
        else:
            self.version = 1

        # Store Pipeline Settings
        self.settings: dict[str, Any] = {"pipeline": init_params}

        # Objective & Scorer
        if self.objective is not None:
            if not isinstance(self.objective, str):
                raise ValueError("Objective needs to be a string.")
            self.scorer = metrics.get_scorer(self.objective)
        else:
            self.scorer = None

        # Required sub-classes
        self.data_sampler = DataSampler()
        self.data_processor = DataProcessor()
        self.data_sequencer = Sequencer()
        self.feature_processor = FeatureProcessor()
        self.interval_analyser = IntervalAnalyser()
        self.drift_detector = DriftDetector()

        # Instance initiating
        self.best_model = None
        self.best_model_str = None
        self.best_feature_set = None
        self.best_score = None
        self._data = None
        self.feature_sets = None
        self.results = None
        self.n_classes = None
        self.is_fitted = False

        # Monitoring
        self._prediction_time = None
        self._main_predictors = None

    # User Pointing Functions
    def get_settings(self, version: int | None = None) -> dict:
        """
        Get settings to recreate fitted object.

        Parameters
        ----------
        version : int, optional
            Production version, defaults to current version
        """
        if version is None or version == self.version:
            assert self.is_fitted, "Pipeline not yet fitted."
            return self.settings
        else:
            settings_path = self.main_dir + f"Production/v{self.version}/Settings.json"
            assert Path(
                settings_path
            ).exists(), "Cannot load settings from nonexistent version"
            return json.load(open(settings_path, "r"))

    def load_settings(self, settings: dict):
        """
        Restores a pipeline from settings.

        Parameters
        ----------
        settings [dict]: Pipeline settings
        """
        # Set parameters
        settings["pipeline"]["no_dirs"] = True
        self.__init__(**settings["pipeline"])
        self.settings = settings
        self.data_processor.load_settings(settings["data_processing"])
        self.feature_processor.load_settings(settings["feature_processing"])
        # TODO: load_settings for IntervalAnalyser (not yet implemented)
        if "drift_detector" in settings:
            self.drift_detector = DriftDetector(
                num_cols=self.data_processor.float_cols + self.data_processor.int_cols,
                cat_cols=self.data_processor.cat_cols,
                date_cols=self.data_processor.date_cols,
            ).load_weights(settings["drift_detector"])

    def load_model(self, model: object):
        """
        Restores a trained model
        """
        assert type(model).__name__ == self.settings["model"]
        self.best_model = model
        self.is_fitted = True

    def fit(
        self,
        data_or_path: np.ndarray | pd.DataFrame | str | Path,
        target: np.ndarray | pd.Series | str | None = None,
        *,
        metadata: dict[int, dict[str, str | float]] | None = None,
        model: str | list[str] | None = None,
        feature_set: str | list[str] | None = None,
        parameter_set: dict | None = None,
        params: dict | None = None,
    ):
        """
        Fit the full AutoML pipeline.
            1. Prepare data for training
            2. Train / optimize models
            3. Prepare Production Files
                Nicely organises all required scripts / files to make a prediction

        Parameters
        ----------
        data_or_path : np.ndarray or pd.DataFrame or str or Path
            Data or path to data. Propagated to `self.data_preparation`.
        target : np.ndarray or pd.Series or str
            Target data or column name. Propagated to `self.data_preparation`.
        *
        metadata : dict of {int : dict of {str : str or float}}, optional
            Metadata. Propagated to `self.data_preparation`.
        model : str or list of str, optional
            Constrain grid search and fitting conclusion to given model(s).
            Propagated to `self.model_training` and `self.conclude_fitting`.
        feature_set : str or list of str, optional
            Constrain grid search and fitting conclusion to given feature set(s).
            Propagated to `self.model_training` and `self.conclude_fitting`.
            Options: {rf_threshold, rf_increment, shap_threshold, shap_increment}
        parameter_set : dict, optional
            Parameter grid to optimize over. Propagated to `self.model_training`.
        params : dict, optional
            Constrain parameters for fitting conclusion.
            Propagated to `self.conclude_fitting`.
        """

        # Starting
        self.logger.info(f"\n\n*** Starting Amplo AutoML - {self.name} ***\n\n")

        # Prepare data for training
        self.data_preparation(
            data_or_path=data_or_path, target=target, metadata=metadata
        )

        # Train / optimize models
        self.model_training(
            model=model, feature_set=feature_set, parameter_set=parameter_set
        )

        # Conclude fitting
        self.conclude_fitting(model=model, feature_set=feature_set, params=params)

    def data_preparation(
        self,
        data_or_path: np.ndarray | pd.DataFrame | str | Path,
        target: np.ndarray | pd.Series | str | None = None,
        *,
        metadata: dict[int, dict[str, str | float]] | None = None,
    ):
        """
        Prepare data for modelling
            1. Data Processing
                Cleans all the data. See @DataProcessing
            2. (optional) Exploratory Data Analysis
                Creates a ton of plots which are helpful to improve predictions manually
            3. Feature Processing
                Extracts & Selects. See @FeatureProcessing

        Parameters
        ----------
        data_or_path : np.ndarray or pd.DataFrame or str or Path
            Data or path to data. Propagated to `self._read_data`.
        target : np.ndarray or pd.Series or str
            Target data or column name. Propagated to `self._read_data`.
        *
        metadata : dict of {int : dict of {str : str or float}}, optional
            Metadata. Propagated to `self._read_data`.
        """
        # Reading data
        self._read_data(data_or_path, target, metadata=metadata)

        # Detect mode (classification / regression)
        self._mode_detector()

        # Preprocess Data
        self._data_processing()

        # Fit Drift Detector to input
        num_cols = list(
            set(self.x).intersection(
                self.data_processor.float_cols + self.data_processor.int_cols
            )
        )
        date_cols = list(set(self.x).intersection(self.data_processor.date_cols))
        cat_cols = list(set(self.x) - set(num_cols + date_cols))
        self.drift_detector = DriftDetector(
            num_cols=num_cols, cat_cols=cat_cols, date_cols=date_cols
        )
        self.drift_detector.fit(self.x)

        # Run Exploratory Data Analysis
        self._eda()

        # Balance data
        self._data_sampling()

        # Sequence
        self._sequencing()

        # Interval-analyze data
        self._interval_analysis()

        # Extract and select features
        self._feature_processing()

        # Standardize
        # Standardizing assures equal scales, equal gradients and no clipping.
        # Therefore, it needs to be after sequencing & feature processing, as this
        # alters scales
        self._standardizing()

    def model_training(self, **kwargs):
        """Train models

        1. Initial Modelling
            Runs various models with default parameters for all feature sets
            If Sequencing is enabled, this is where it happens, as here, the feature
            set is generated.
        2. Grid Search
            Optimizes the hyperparameters of the best performing models
        3. (optional) Create Stacking model
        4. (optional) Create documentation

        Parameters
        ----------
        kwargs : optional
            Keyword arguments that will be passed to `self.grid_search`.
        """
        # Run initial models
        self._initial_modelling()

        # Optimize Hyper parameters
        self.grid_search(**kwargs)

        # Create stacking model
        self._create_stacking()

    def conclude_fitting(self, model=None, feature_set=None, params=None):
        """
        Prepare production files that are necessary to deploy a specific
        model / feature set combination

        Creates or modifies the following files
            - ``Model.joblib`` (production model)
            - ``Settings.json`` (model settings)

        Parameters
        ----------
        model : str or list of str, optional
            Model file for which to prepare production files.
        feature_set : str or list of str, optional
            Feature set for which to prepare production files.
        params : dict, optional
            Model parameters for which to prepare production files.
            Default: takes the best parameters
        """

        if not self.no_dirs:

            # Set up production path
            prod_dir = self.main_dir + f"Production/v{self.version}/"
            Path(prod_dir).mkdir(exist_ok=True)

            # Parse arguments
            self._parse_production_args(model, feature_set, params)

            # Verbose printing
            self.logger.info(
                f"Preparing Production files for {self.best_model_str}, "
                f"{self.best_feature_set}"
            )

            # Set best model
            self._prepare_production_model(prod_dir + "Model.joblib")

            # Set and store production settings
            self._prepare_production_settings(prod_dir + "Settings.json")

            # Check data
            obs = DataObserver(pipeline=self)
            obs.observe()
            self.settings["data_observer"] = obs.observations

            # Observe production
            obs = ModelObserver(pipeline=self)
            obs.observe()
            self.settings["model_observer"] = obs.observations

        # Finish
        self.is_fitted = True
        self.logger.info("All done :)")
        utils.logging.del_file_handlers()

    def convert_data(
        self, x: pd.DataFrame, preprocess: bool = True
    ) -> Tuple[pd.DataFrame, pd.Series]:
        """
        Function that uses the same process as the pipeline to clean data.
        Useful if pipeline is pickled for production

        Parameters
        ----------
        data [pd.DataFrame]: Input features
        """
        # Convert to Pandas
        if isinstance(x, np.ndarray):
            x = pd.DataFrame(x, columns=[f"Feature_{i}" for i in range(x.shape[1])])

        # Custom code
        if self.preprocess_function is not None and preprocess:
            ex_globals = {"data": x}
            exec(self.preprocess_function, ex_globals)
            x = ex_globals["data"]

        # Process data
        x = self.data_processor.transform(x)

        # Drift Check
        self.drift_detector.check(x)

        # Split output
        y = None
        if self.target in x.keys():
            y = x[self.target]
            if not self.include_output:
                x = x.drop(self.target, axis=1)

        # Sequence
        if self.sequence:
            x, y = self.data_sequencer.convert(x, y)

        # Convert Features
        x = self.feature_processor.transform(
            x, feature_set=self.settings["feature_set"]
        )

        # Standardize
        if self.standardize:
            x, y = self._transform_standardize(x, y)

        # NaN test -- datetime should be taken care of by now
        if (
            x.astype("float32").replace([np.inf, -np.inf], np.nan).isna().sum().sum()
            != 0
        ):
            raise ValueError(
                f"Column(s) with NaN: {list(x.keys()[x.isna().sum() > 0])}"
            )

        # Return
        return x, y

    def predict(self, data: pd.DataFrame) -> np.ndarray:
        """
        Full script to make predictions. Uses 'Production' folder with defined or
        latest version.

        Parameters
        ----------
        data [pd.DataFrame]: data to do prediction on
        """
        start_time = time.time()
        assert self.is_fitted, "Pipeline not yet fitted."

        # Print
        self.logger.info(
            f"Predicting with {type(self.best_model).__name__}, v{self.version}"
        )

        # Convert
        x, y = self.convert_data(data)

        # Predict
        predictions = self.best_model.predict(x)
        if self.mode == "regression" and self.standardize:
            predictions = self._inverse_standardize(predictions)
        elif self.mode == "classification":
            predictions[:] = self.data_processor.decode_labels(
                predictions.astype(int), except_not_fitted=False
            )

        # Stop timer
        self._prediction_time = (time.time() - start_time) / len(x) * 1000

        # Calculate main predictors
        self._get_main_predictors(x)

        return predictions

    def predict_proba(self, data: pd.DataFrame) -> np.ndarray:
        """
        Returns probabilistic prediction, only for classification.

        Parameters
        ----------
        data [pd.DataFrame]: data to do prediction on
        """
        start_time = time.time()
        assert self.is_fitted, "Pipeline not yet fitted."
        assert (
            self.mode == "classification"
        ), "Predict_proba only available for classification"
        assert hasattr(
            self.best_model, "predict_proba"
        ), f"{type(self.best_model).__name__} has no attribute predict_proba"

        # Logging
        self.logger.info(
            f"Predicting with {type(self.best_model).__name__}, v{self.version}"
        )

        # Convert data
        x, y = self.convert_data(data)

        # Predict
        prediction = self.best_model.predict_proba(x)

        # Stop timer
        self._prediction_time = (time.time() - start_time) / len(x) * 1000

        # Calculate main predictors
        self._get_main_predictors(x)

        return prediction

    # Fit functions
    def _read_data(
        self,
        data_or_path: np.ndarray | pd.DataFrame | str | Path,
        target: np.ndarray | pd.Series | str | None = None,
        *,
        metadata: dict[int, dict[str, str | float]] | None = None,
    ) -> "Pipeline":
        """
        Read and validate data.

        Notes
        -----
        The required parameters depend on the input parameter types.

        When ``target`` is None, it is set to ``self.target`` or "target" otherwise.

        When ``data_or_path`` is path-like, then the parameters ``target`` and
        ``metadata`` are not required.
        Otherwise, when ``data_or_path`` is array-like, it either must contain a column
        name as the ``target`` parameter indicates or ``target`` must also be an
        array-like object with the same length as ``data_or_path``.

        Parameters
        ----------
        data_or_path : np.ndarray or pd.DataFrame or str or Path
            Data or path to data.
        target : np.ndarray or pd.Series or str
            Target data or column name.
        *
        metadata : dict of {int : dict of {str : str or float}}, optional
            Metadata.

        Returns
        -------
        Pipeline
            The same object but with injected data.
        """

        # Allow target name to be set via __init__
        target_name = utils.clean_feature_name(
            (target if isinstance(target, str) else None) or self.target or "target"
        )

        # Read / set data
        if isinstance(data_or_path, (str, Path)):
            if not isinstance(target, (type(None), str)):
                raise ValueError(
                    "Parameter `target` must be a string when `data_or_path` is a "
                    "path-like object."
                )
            if metadata:
                warn(
                    "Parameter `metadata` is ignored when `data_or_path` is a "
                    "path-like object."
                )

            data, metadata = utils.io.merge_logs(data_or_path, target_name)

        elif isinstance(data_or_path, (np.ndarray, pd.DataFrame)):
            data = pd.DataFrame(data_or_path)

        else:
            raise ValueError(f"Invalid type for `data_or_path`: {type(data_or_path)}")

        data = utils.data.clean_keys(data)

        # Validate target
        if target is None or isinstance(target, str):
            if target_name not in data:
                raise ValueError(f"Target column '{target_name}' not found in data.")

        elif isinstance(target, (np.ndarray, pd.Series)):

            if len(data) != len(target):
                raise ValueError("Length of target and data don't match.")
            elif not isinstance(target, pd.Series):
                target = pd.Series(target, index=data.index)
            elif not all(data.index == target.index):
                warn(
                    "Indices of data and target don't match. Target index will be "
                    "overwritten by data index."
                )
                target.index = data.index

            if target_name in data:
                raise ValueError(f"A '{target_name}' column already exists in `data`.")
            data[target_name] = target

        else:
            raise ValueError("Invalid type for `target`.")

        assert target_name in data, "Internal error: Target not in data."
        self.target = target_name

        # Finish
        self._set_data(data)
        self.settings["file_metadata"] = metadata or {}

        return self

    def has_new_training_data(self):
        # Return True if no previous version exists
        if self.version == 1:
            return True

        # Get previous and current file metadata
        curr_metadata = self.settings["file_metadata"]
        last_metadata = self.get_settings(self.version - 1)["file_metadata"]

        # Check each settings file
        for file_id in curr_metadata:
            # Get file specific metadata
            curr = curr_metadata[file_id]
            last = last_metadata.get(file_id, dict())
            # Compare metadata
            same_folder = curr["folder"] == last.get("folder")
            same_file = curr["file"] == last.get("file")
            same_mtime = curr["last_modified"] == last.get("last_modified")
            if not all([same_folder, same_file, same_mtime]):
                return False

        return True

    def _mode_detector(self):
        """
        Detects the mode (Regression / Classification)
        """
        # Only run if mode is not provided
        if self.mode is not None:
            return

        # Classification if string
        if self.y.dtype == str or self.y.nunique() < 0.1 * len(self.data):
            self.mode = "classification"
            self.objective = self.objective or "neg_log_loss"

        # Else regression
        else:
            self.mode = "regression"
            self.objective = self.objective or "neg_mean_absolute_error"

        # Set scorer
        self.scorer = metrics.get_scorer(self.objective)

        # Copy to settings
        self.settings["pipeline"]["mode"] = self.mode
        self.settings["pipeline"]["objective"] = self.objective

        # Logging
        self.logger.info(
            f"Setting mode to {self.mode} & objective to {self.objective}."
        )

    def _data_processing(self):
        """
        Organises the data cleaning. Heavy lifting is done in self.dataProcessor,
        but settings etc. needs to be organised.
        """
        self.data_processor = DataProcessor(
            target=self.target,
            int_cols=self.int_cols,
            float_cols=self.float_cols,
            date_cols=self.date_cols,
            cat_cols=self.cat_cols,
            include_output=True,
            missing_values=self.missing_values,
            outlier_removal=self.outlier_removal,
            z_score_threshold=self.z_score_threshold,
        )

        # Set paths
        extracted_data_path = self.main_dir + f"Data/Extracted_v{self.version}.csv"
        data_path = self.main_dir + f"Data/Cleaned_v{self.version}.csv"
        settings_path = self.main_dir + f"Settings/Cleaning_v{self.version}.json"

        # Only load settings if feature processor is done already.
        if Path(extracted_data_path).exists() and Path(settings_path).exists():
            self.settings["data_processing"] = json.load(open(settings_path, "r"))
            self.data_processor.load_settings(self.settings["data_processing"])

        # Else, if data processor is done, load settings & data
        elif Path(data_path).exists() and Path(settings_path).exists():
            # Load data
            data = self._read_csv(data_path)
            self._set_data(data)

            # Load settings
            self.settings["data_processing"] = json.load(open(settings_path, "r"))
            self.data_processor.load_settings(self.settings["data_processing"])

            self.logger.info("Loaded Cleaned Data")

        # Else, run the data processor
        else:
            # Cleaning
            data = self.data_processor.fit_transform(self.data)

            # Update pipeline
            self._set_data(data)

            # Store data
            self._write_csv(self.data, data_path)

            # Clean up Extracted previous version
            if os.path.exists(
                self.main_dir + f"Data/Extracted_v{self.version - 1}.csv"
            ):
                os.remove(self.main_dir + f"Data/Extracted_v{self.version - 1}.csv")

            # Save settings
            self.settings["data_processing"] = self.data_processor.get_settings()
            if not self.no_dirs:
                json.dump(self.settings["data_processing"], open(settings_path, "w"))

        # If no columns were provided, load them from data processor
        if self.date_cols is None:
            self.date_cols = self.settings["data_processing"]["date_cols"]
        if self.int_cols is None:
            self.int_cols = self.settings["data_processing"]["int_cols"]
        if self.float_cols is None:
            self.float_cols = self.settings["data_processing"]["float_cols"]
        if self.cat_cols is None:
            self.cat_cols = self.settings["data_processing"]["cat_cols"]

        # Assert classes in case of classification
        if self.mode == "classification":
            self.n_classes = self.y.nunique()
            if self.n_classes >= 50:
                warn(
                    "More than 50 classes, "
                    "you may want to reconsider classification mode"
                )
            if set(self.y) != set([i for i in range(len(set(self.y)))]):
                raise ValueError("Classes should be [0, 1, ...]")

    def _eda(self):
        if not self.plot_eda:
            return

        self.logger.info("Starting Exploratory Data Analysis")
        eda = DataExplorer(
            self.x,
            y=self.y,
            mode=self.mode,
            folder=self.main_dir,
            version=self.version,
        )
        eda.run()

    def _data_sampling(self):
        """
        Only run for classification problems. Balances the data using imblearn.
        Does not guarantee to return balanced classes. (Methods are data dependent)
        """
        if not self.balance or not self.mode == "classification":
            return

        self.data_sampler = DataSampler(
            method="both",
            margin=0.1,
            cv_splits=self.cv_splits,
            shuffle=self.cv_shuffle,
            fast_run=False,
            objective=self.objective,
        )
        data_path = self.main_dir + f"Data/Balanced_v{self.version}.csv"

        if Path(data_path).exists():
            # Load data
            data = self._read_csv(data_path)
            self._set_data(data)

            self.logger.info("Loaded Balanced data")

        else:
            # Fit and resample
            self.logger.info("Resampling data")
            x, y = self.data_sampler.fit_resample(self.x, self.y)

            # Store
            self._set_xy(x, y)
            self._write_csv(self.data, data_path)

    def _sequencing(self):
        """
        Sequences the data. Useful mostly for problems where older samples play a role
        in future values. The settings of this module are NOT AUTOMATIC
        """
        if not self.sequence:
            return

        self.data_sequencer = Sequencer(
            back=self.sequence_back,
            forward=self.sequence_forward,
            shift=self.sequence_shift,
            diff=self.sequence_diff,
        )
        data_path = self.main_dir + f"Data/Sequence_v{self.version}.csv"

        if Path(data_path).exists():
            # Load data
            data = self._read_csv(data_path)
            self._set_data(data)

            self.logger.info("Loaded Extracted Features")

        else:
            # Sequencing
            self.logger.info("Sequencing data")
            x, y = self.data_sequencer.convert(self.x, self.y)

            # Store
            self._set_xy(x, y)
            self._write_csv(self.data, data_path)

    def _feature_processing(self):
        """
        Organises feature processing. Heavy lifting is done in self.featureProcessor,
        but settings, etc. needs to be organised.
        """
        self.feature_processor = FeatureProcessor(
            mode=self.mode,
            is_temporal=None,
            extract_features=self.extract_features,
            collinear_threshold=self.information_threshold,
            verbose=self.verbose,
        )

        # Set paths
        data_path = self.main_dir + f"Data/Extracted_v{self.version}.csv"
        settings_path = self.main_dir + f"Settings/Extracting_v{self.version}.json"

        if Path(data_path).exists() and Path(settings_path).exists():
            # Loading data
            data = self._read_csv(data_path)
            self._set_data(data)

            # Loading settings
            self.settings["feature_processing"] = json.load(open(settings_path, "r"))
            self.feature_processor.load_settings(self.settings["feature_processing"])
            self.feature_sets = self.settings["feature_processing"]["feature_sets_"]

            self.logger.info("Loaded Extracted Features")

        else:
            self.logger.info("Starting Feature Processor")

            # Transform data.  Note that y also needs to be transformed in the
            # case when we're using the temporal feature processor (pooling).
            x = self.feature_processor.fit_transform(self.x, self.y)
            y = self.feature_processor.transform_target(self.y)
            self.feature_sets = self.feature_processor.feature_sets_

            # Store data
            self._set_xy(x, y)
            self._write_csv(self.data, data_path)

            # Cleanup Cleaned_vx.csv
            if os.path.exists(self.main_dir + f"Data/Cleaned_v{self.version}.csv"):
                os.remove(self.main_dir + f"Data/Cleaned_v{self.version}.csv")

            # Save settings
            self.settings["feature_processing"] = self.feature_processor.get_settings()
            if not self.no_dirs:
                json.dump(self.settings["feature_processing"], open(settings_path, "w"))

    def _interval_analysis(self):
        """
        Interval-analyzes the data using ``amplo.auto_ml.interval_analysis``
        or resorts to pre-computed data, if present.
        """
        # Skip analysis when analysis is not possible and/or not desired
        is_multi_index = len(self.x.index.names) == 2
        if not all(
            [self.use_interval_analyser, is_multi_index, self.mode == "classification"]
        ):
            return

        self.interval_analyser = IntervalAnalyser(target=self.target)

        # Set paths
        data_path = self.main_dir + f"Data/Interval_Analyzed_v{self.version}.csv"
        # settings_path = self.main_dir +
        # f'Settings/Interval_Analysis_v{self.version}.json'

        if Path(data_path).exists():  # TODO: and Path(settings_path).exists():
            # Load data
            data = self._read_csv(data_path)
            self._set_data(data)

            # TODO implement `IntervalAnalyser.load_settings` and add to
            #  `self.load_settings`
            # # Load settings
            # self.settings['interval_analysis'] = json.load(open(settings_path, 'r'))
            # self.intervalAnalyser.load_settings(self.settings['interval_analysis'])

            self.logger.info("Loaded interval-analyzed data")

        else:
            self.logger.info("Interval-analyzing data")

            # Transform data
            data = self.interval_analyser.fit_transform(self.x, self.y)

            # Store data
            self._set_data(data)
            self._write_csv(self.data, data_path)

            # TODO implement `IntervalAnalyser.get_settings` and add to
            #  `self.get_settings`
            # # Save settings
            # self.settings['interval_analysis'] = self.intervalAnalyser.get_settings()
            # json.dump(self.settings['interval_analysis'], open(settings_path, 'w'))

    def _standardizing(self):
        """
        Wrapper function to determine whether to fit or load
        """
        # Return if standardize is off
        if not self.standardize:
            return

        # Set paths
        settings_path = self.main_dir + f"Settings/Standardize_v{self.version}.json"

        if Path(settings_path).exists():
            # Load data
            self.settings["standardize"] = json.load(open(settings_path, "r"))

        else:
            # Fit data
            self._fit_standardize(self.x, self.y)

            # Store Settings
            json.dump(self.settings["standardize"], open(settings_path, "w"))

        # Transform data
        x, y = self._transform_standardize(self.x, self.y)
        self._set_xy(x, y)

    def _initial_modelling(self):
        """
        Runs various models to see which work well.
        """

        # Set paths
        results_path = Path(self.main_dir) / "Results.csv"

        # Load existing results
        if results_path.exists():

            # Load results
            self.results = pd.read_csv(results_path)

            # Printing here as we load it
            results = self.results[
                np.logical_and(
                    self.results["version"] == self.version,
                    self.results["type"] == "Initial modelling",
                )
            ]
            for fs in set(results["dataset"]):
                self.logger.info(
                    f"Initial Modelling for {fs} ({len(self.feature_sets[fs])})"
                )
                fsr = results[results["dataset"] == fs]
                for i in range(len(fsr)):
                    row = fsr.iloc[i]
                    self.logger.info(
                        f'{row["model"].ljust(40)} {self.objective}: '
                        f'{row["mean_objective"]:.4f} \u00B1 {row["std_objective"]:.4f}'
                    )

        # Check if this version has been modelled
        if self.results is None or self.version not in self.results["version"].values:

            # Iterate through feature sets
            for feature_set, cols in self.feature_sets.items():

                # Skip empty sets
                if len(cols) == 0:
                    self.logger.info(f"Skipping {feature_set} features, empty set")
                    continue
                self.logger.info(
                    f"Initial Modelling for {feature_set} features ({len(cols)})"
                )

                # Do the modelling
                modeller = Modeller(
                    mode=self.mode,
                    store_models=self.store_models,
                    cv=self.cv,
                    objective=self.objective,
                    dataset=feature_set,
                    store_results=False,
                    folder=self.main_dir + "Models/",
                )
                results = modeller.fit(self.x[cols], self.y)

                # Add results to memory
                results["type"] = "Initial modelling"
                results["version"] = self.version
                if self.results is None:
                    self.results = results
                else:
                    self.results = pd.concat([self.results, results])

            # Save results
            self._write_csv(self.results, results_path)

    def grid_search(self, model=None, feature_set=None, parameter_set=None):
        """
        Runs a grid search.

        By default, takes ``self.results`` and runs for the top ``n =
        self.n_grid_searches`` optimizations. There is the option to provide ``model``
        and ``feature_set``, but **both** have to be provided. In this case, the model
        and dataset combination will be optimized.

        Implemented types, Exhaustive, Halving, Optuna.

        Parameters
        ----------
        model : list of (str or object) or object or str, optional
            Which model to run grid search for.
        feature_set : list of str or str, optional
            Which feature set to run gid search for. Must be provided when `model` is
            not None.
            Options: {rf_threshold, rf_increment, shap_threshold, shap_increment}
        parameter_set : dict, optional
            Parameter grid to optimize over.

        Notes
        -----
        When both parameters, ``model`` and ``feature_set``, are provided, the grid
        search behaves as follows:
            - When both parameters are either of dtype ``str`` or have the same length,
              then grid search will treat them as pairs.
            - When one parameter is an iterable and the other parameter is either a
              string or an iterable of different length, then grid search will happen
              for each unique combination of these parameters.
        """

        # Skip grid search and set best initial model as best grid search parameters
        if self.grid_search_type is None or self.n_grid_searches == 0:
            best_initial_model = self._sort_results(
                self.results[self.results["version"] == self.version]
            ).iloc[:1]
            best_initial_model["type"] = "Hyper Parameter"
            self.results = pd.concat(
                [self.results, best_initial_model], ignore_index=True
            )
            return self

        # Define models
        if model is None:
            # Run through first best initial models (n = self.n_grid_searches)
            selected_results = self.sort_results(
                self.results[
                    np.logical_and(
                        self.results["type"] == "Initial modelling",
                        self.results["version"] == self.version,
                    )
                ]
            ).iloc[: self.n_grid_searches]
            models = [
                utils.get_model(model_name, mode=self.mode, samples=len(self.x))
                for model_name in selected_results["model"]
            ]
            feature_sets = selected_results["dataset"]

        elif feature_set is None:
            raise AttributeError(
                "When `model` is provided, `feature_set` cannot be None. "
                "Provide either both params or neither of them."
            )

        else:
            models = (
                [utils.get_model(model, mode=self.mode, samples=len(self.x))]
                if isinstance(model, str)
                else [model]
            )
            feature_sets = (
                [feature_set] if isinstance(feature_set, str) else list(feature_set)
            )
            if len(models) != len(feature_sets):
                # Create each combination
                combinations = list(
                    itertools.product(np.unique(models), np.unique(feature_sets))
                )
                models = [elem[0] for elem in combinations]
                feature_sets = [elem[1] for elem in combinations]

        # Iterate and grid search over each pair of model and feature_set
        for model, feature_set in zip(models, feature_sets):

            # Organise existing model results
            m_results = self.results[
                np.logical_and(
                    self.results["model"] == type(model).__name__,
                    self.results["version"] == self.version,
                )
            ]
            m_results = self._sort_results(
                m_results[m_results["dataset"] == feature_set]
            )

            # Skip grid search if optimized model already exists
            if not ("Hyper Parameter" == m_results["type"]).any():
                # Run grid search for model
                grid_search_results = self._grid_search_iteration(
                    model, parameter_set, feature_set
                )
                grid_search_results = self.sort_results(grid_search_results)

                # Store results
                grid_search_results["version"] = self.version
                grid_search_results["dataset"] = feature_set
                grid_search_results["type"] = "Hyper Parameter"
                self.results = pd.concat(
                    [self.results, grid_search_results], ignore_index=True
                )
                self.results.to_csv(self.main_dir + "Results.csv", index=False)

        return self

    def _create_stacking(self):
        """
        Based on the best performing models, in addition to cheap models based on very
        different assumptions, a stacking model is optimized to enhance/combine the
        performance of the models.
        --> should contain a large variety of models
        --> classifiers need predict_proba
        --> level 1 needs to be ordinary least squares
        """
        if not self.stacking:
            return

        self.logger.info("Creating Stacking Ensemble")

        # Select most frequent feature set from hyperparameter optimization
        results = self._sort_results(
            self.results[
                np.logical_and(
                    self.results["type"] == "Hyper Parameter",
                    self.results["version"] == self.version,
                )
            ]
        )
        feature_set = results["dataset"].value_counts().index[0]
        results = results[results["dataset"] == feature_set]
        self.logger.info("Selected Stacking feature set: {}".format(feature_set))

        # Create Stacking Model Params
        n_stacking_models = 3
        stacking_models_str = results["model"].unique()[:n_stacking_models]
        stacking_models_params = [
            utils.io.parse_json(
                results.iloc[np.where(results["model"] == sms)[0][0]]["params"]
            )
            for sms in stacking_models_str
        ]
        stacking_models = dict(
            [
                (sms, stacking_models_params[i])
                for i, sms in enumerate(stacking_models_str)
            ]
        )
        self.logger.info("Stacked models: {}".format(list(stacking_models.keys())))
        # Make the dict of dict flat
        add_to_stack = list(stacking_models)
        stacking_models = {
            f"{model}__{param}": parameters[param]
            for model, parameters in stacking_models.items()
            for param in parameters
        }

        # Add samples & Features
        stacking_models["n_samples"], stacking_models["n_features"] = self.x.shape

        # Prepare Stack
        if self.mode == "regression":
            stack = StackingRegressor(add_to_stack, **stacking_models)
        elif self.mode == "classification":
            stack = StackingClassifier(add_to_stack, **stacking_models)
        else:
            raise NotImplementedError("Unknown mode")

        # Cross Validate
        x, y = self.x[self.feature_sets[feature_set]].to_numpy(), self.y.to_numpy()
        score = []
        times = []
        for (t, v) in tqdm(self.cv.split(x, y)):
            start_time = time.time()
            xt, xv, yt, yv = x[t], x[v], y[t].reshape((-1)), y[v].reshape((-1))
            model = copy.deepcopy(stack)
            model.fit(xt, yt)
            score.append(self.scorer(model, xv, yv))
            times.append(time.time() - start_time)

        # Output Results
        self.logger.info("Stacking result:")
        self.logger.info(
            f"{self.objective}:     {np.mean(score):.2f} \u00B1 {np.std(score):.2f}"
        )
        self.results = self.results.append(
            {
                "date": datetime.today().strftime("%d %b %y"),
                "model": type(stack).__name__,
                "dataset": feature_set,
                "params": json.dumps(stack.get_params()),
                "mean_objective": np.mean(score),
                "std_objective": np.std(score),
                "mean_time": np.mean(times),
                "std_time": np.std(times),
                "version": self.version,
                "type": "Stacking",
            },
            ignore_index=True,
        )
        self.results.to_csv(self.main_dir + "Results.csv", index=False)

    def _parse_production_args(self, model=None, feature_set=None, params=None):
        """
        Parse production arguments. Selects the best model, feature set and parameter
        combination.

        Parameters
        ----------
        model : str or list of str, optional
            Model constraint(s). In case list is provided, the pipeline needs to be
            fitted.
        feature_set : str or list of str, optional
            Feature set constraint(s). In case list is provided, the pipeline needs to
            be fitted.
        params : dict, optional
            Parameter constraint(s)

        Returns
        -------
        model : str
            Best model given the `model` restraint(s).
        feature_set : str
            Best feature set given the `feature_set` restraint(s).
        params : dict
            Best model parameters given the `params` restraint(s).
        """
        if self.results is None and (
            model is None or params is None or feature_set is None
        ):
            raise ValueError(
                "Pipeline not fitted and no model, params or feature set provided."
            )

        if model is not None and not isinstance(model, str):
            # TODO: This issue is linked with AML-103 (in Jira)
            #  1. Add to method docstring that it accepts a model instance, too
            #  2. Change `if`-case to a single `isinstance(model, BasePredictor)`
            # Get model name
            model = type(model).__name__

        # Get results of current version
        results = (
            self._sort_results(self.results[self.results["version"] == self.version])
            if self.results is not None
            else None
        )

        # Filter results for model
        if model is not None and results is not None:
            # Enforce list
            if isinstance(model, str):
                model = [model]

            # Filter results
            results = self._sort_results(results[results["model"].isin(model)])

        # filter results for feature set
        if feature_set is not None and results is not None:
            if isinstance(feature_set, str):
                feature_set = [feature_set]
            # Filter results
            results = self._sort_results(results[results["dataset"].isin(feature_set)])

        # Get best parameters
        if params is None and results is not None:
            params = results.iloc[0]["params"]
        elif params is None:
            params = {}

        # Find the best allowed arguments
        self.best_model_str = model if results is None else results.iloc[0]["model"]
        self.best_feature_set = (
            feature_set if results is None else results.iloc[0]["dataset"]
        )
        self.best_params = utils.io.parse_json(params)
        self.best_score = 0 if results is None else results.iloc[0]["worst_case"]

        return self

    def _prepare_production_model(self, model_path):
        """
        Prepare and store `self.bestModel` for production

        Parameters
        ----------
        model_path : str or Path
            Where to store model for production

        Returns
        -------
        """
        model_path = Path(model_path)

        # Make model
        if "Stacking" in self.best_model_str:
            # Create stacking
            if self.mode == "regression":
                stacking_model = StackingRegressor
            elif self.mode == "classification":
                stacking_model = StackingClassifier
            else:
                raise NotImplementedError("Mode not set")

            self.best_model = stacking_model(
                add_to_stack=self.best_params.get("add_to_stack", None),
                add_defaults_to_stack=self.best_params.get(
                    "add_defaults_to_stack", True
                ),
                n_samples=self.x.shape[0],
                n_features=self.x.shape[1],
            )
        else:
            # Take model as is
            self.best_model = utils.get_model(
                self.best_model_str, mode=self.mode, samples=len(self.x)
            )

        # Set params, train
        self.best_model.set_params(**self.best_params)
        self.best_model.fit(self.x[self.feature_sets[self.best_feature_set]], self.y)

        # Save model
        joblib.dump(self.best_model, model_path)

        self.best_score = self.scorer(
            self.best_model,
            self.x[self.feature_sets[self.best_feature_set]],
            self.y,
        )
        self.logger.info(
            f"Model fully fitted, in-sample {self.objective}: {self.best_score:4f}"
        )

        return

    def _prepare_production_settings(self, settings_path):
        """
        Prepare `self.settings` for production and dump to file

        Parameters
        ----------
        settings_path : str or Path
            Where to save settings for production
        """
        assert self.best_model is not None, "`self.bestModel` is not yet prepared"
        settings_path = Path(settings_path)

        # Update pipeline settings
        self.settings["version"] = self.version
        self.settings["pipeline"]["verbose"] = self.verbose
        self.settings["model"] = self.best_model_str
        self.settings["params"] = self.best_params
        self.settings["feature_set"] = self.best_feature_set
        self.settings["features"] = self.feature_sets[self.best_feature_set]
        self.settings["best_score"] = self.best_score
        self.settings["amplo_version"] = (
            amplo.__version__ if hasattr(amplo, "__version__") else "dev"
        )

        # Validation
        validator = ModelValidator(
            cv_splits=self.cv_splits,
            cv_shuffle=self.cv_shuffle,
            verbose=self.verbose,
        )
        self.settings["validation"] = validator.validate(
            model=self.best_model, x=self.x, y=self.y, mode=self.mode
        )

        # Prune Data Processor
        required_features = get_required_columns(
            self.feature_processor.feature_sets_[self.best_feature_set],
            self.feature_processor.numeric_cols_,
        )
        self.data_processor.prune_features(required_features)
        self.settings["data_processing"] = self.data_processor.get_settings()

        # Fit Drift Detector to output and store settings
        self.drift_detector.fit_output(
            self.best_model, self.x[self.feature_sets[self.best_feature_set]]
        )
        self.settings["drift_detector"] = self.drift_detector.get_weights()

        # Save settings
        json.dump(self.settings, open(settings_path, "w"), indent=4)

    # Getter Functions / Properties
    @property
    def cv(self):
        """
        Gives the Cross Validation scheme

        Returns
        -------
        cv : KFold or StratifiedKFold
            The cross validator
        """
        # Regression
        if self.mode == "regression":
            return KFold(
                n_splits=self.cv_splits,
                shuffle=self.cv_shuffle,
                random_state=83847939 if self.cv_shuffle else None,
            )

        # Classification
        if self.mode == "classification":
            return StratifiedKFold(
                n_splits=self.cv_splits,
                shuffle=self.cv_shuffle,
                random_state=83847939 if self.cv_shuffle else None,
            )

    @property
    def data(self) -> Union[None, pd.DataFrame]:
        return self._data

    @property
    def x(self) -> pd.DataFrame:
        if self.data is None:
            raise AttributeError("Data is None")
        if self.include_output:
            return self.data
        return self.data.drop(self.target, axis=1)

    @property
    def y(self):
        if self.data is None:
            raise AssertionError("`self.data` is empty. Set a value with `set_data`")
        return self.data[self.target]

    @property
    def y_orig(self):
        enc_labels = self.y
        dec_labels = self.data_processor.decode_labels(
            enc_labels, except_not_fitted=False
        )
        return pd.Series(dec_labels, name=self.target, index=enc_labels.index)

    # Setter Functions
    def _set_data(self, new_data: pd.DataFrame):
        assert isinstance(new_data, pd.DataFrame), "Invalid data type"
        assert self.target in new_data, "No target column present"
        assert len(new_data.columns) > 1, "No feature column present"
        self._data = new_data

    def _set_xy(
        self,
        new_x: Union[np.ndarray, pd.DataFrame],
        new_y: Union[np.ndarray, pd.Series],
    ):
        if not isinstance(new_y, pd.Series):
            new_y = pd.Series(new_y, name=self.target)
        new_data = pd.concat([new_x, new_y], axis=1)
        self._set_data(new_data)

    def _set_x(self, new_x: Union[np.ndarray, pd.DataFrame]):
        # Convert to dataframe
        if isinstance(new_x, np.ndarray):
            old_x = self.data.drop(self.target, axis=1)
            old_x_shape = old_x.shape
            old_x_index = old_x.index
            old_x_columns = old_x.columns
            del old_x

            if new_x.shape == old_x_shape:
                new_x = pd.DataFrame(new_x, index=old_x_index, columns=old_x_columns)
            else:
                warn(
                    "Old x-data has more/less columns than new x-data. "
                    "Setting dummy feature names..."
                )
                columns = [f"Feature_{i}" for i in range(new_x.shape[1])]
                new_x = pd.DataFrame(new_x, index=old_x_index, columns=columns)

        elif not isinstance(new_x, pd.DataFrame):
            raise ValueError(f"Invalid dtype for new x data: {type(new_x)}")

        # Assert that target is not in x-data
        if self.target in new_x:
            raise AttributeError("Target column name should not be in x-data")

        # Set data
        self._data = pd.concat([new_x, self.y], axis=1)

    def set_y(self, new_y: Union[np.ndarray, pd.Series]):
        self._data[self.target] = new_y

    # Support Functions
    @staticmethod
    def _read_csv(data_path) -> pd.DataFrame:
        """
        Read data from given path and set index or multi-index

        Parameters
        ----------
        data_path : str or Path
        """
        assert Path(data_path).suffix == ".csv", "Expected a *.csv path"

        data = pd.read_csv(data_path)

        if {"index", "log"}.issubset(data.columns):
            # Multi-index: case when IntervalAnalyser was used
            index = ["log", "index"]
        elif "index" in data.columns:
            index = ["index"]
        else:
            raise IndexError(
                "No known index was found. "
                "Expected to find at least a column named `index`."
            )
        return data.set_index(index)

    def _write_csv(self, data, data_path):
        """
        Write data to given path and set index if needed.

        Parameters
        ----------
        data : pd.DataFrame or pd.Series
        data_path : str or Path
        """
        assert Path(data_path).suffix == ".csv", "Expected a *.csv path"

        # Set single-index if not already present
        if len(data.index.names) == 1 and data.index.name is None:
            data.index.name = "index"

        # Raise error if unnamed index is present
        if None in data.index.names:
            raise IndexError(
                f"Found an unnamed index column ({list(data.index.names)})."
            )

        # Write data
        if not self.no_dirs:
            data.to_csv(data_path)

    def _load_version(self):
        """
        Upon start, loads version
        """
        # No need if version is set
        if self.version is not None:
            return

        # Find production folders
        completed_versions = len(os.listdir(self.main_dir + "Production"))
        self.version = completed_versions + 1

        self.logger.info(f"Setting Version {self.version}")

    def _create_dirs(self):
        folders = [
            "",
            "EDA",
            "Data",
            "Features",
            "Documentation",
            "Production",
            "Settings",
        ]
        for folder in folders:
            if not os.path.exists(self.main_dir + folder):
                os.makedirs(self.main_dir + folder)

    def sort_results(self, results: pd.DataFrame) -> pd.DataFrame:
        return self._sort_results(results)

    def _fit_standardize(self, x: pd.DataFrame, y: pd.Series):
        """
        Fits a standardization parameters and returns the transformed data
        """

        # Fit Input
        cat_cols = [
            k
            for lst in self.settings["data_processing"]["dummies"].values()
            for k in lst
        ]
        features = [
            k for k in x.keys() if k not in self.date_cols and k not in cat_cols
        ]
        means_ = x[features].mean(axis=0)
        stds_ = x[features].std(axis=0)
        stds_[stds_ == 0] = 1
        settings = {
            "input": {
                "features": features,
                "means": means_.to_list(),
                "stds": stds_.to_list(),
            }
        }

        # Fit Output
        if self.mode == "regression":
            std = y.std()
            settings["output"] = {
                "mean": y.mean(),
                "std": std if std != 0 else 1,
            }

        self.settings["standardize"] = settings

    def _transform_standardize(
        self, x: pd.DataFrame, y: pd.Series
    ) -> Tuple[pd.DataFrame, pd.Series]:
        """
        Standardizes the input and output with values from settings.

        Parameters
        ----------
        x [pd.DataFrame]: Input data
        y [pd.Series]: Output data
        """
        # Input
        assert self.settings["standardize"], "Standardize settings not found."

        # Pull from settings
        features = self.settings["standardize"]["input"]["features"]
        means = self.settings["standardize"]["input"]["means"]
        stds = self.settings["standardize"]["input"]["stds"]

        # Filter if not all features are present
        if len(x.keys()) < len(features):
            indices = [
                [i for i, j in enumerate(features) if j == k][0] for k in x.keys()
            ]
            features = [features[i] for i in indices]
            means = [means[i] for i in indices]
            stds = [stds[i] for i in indices]

        # Transform Input
        x[features] = (x[features] - means) / stds

        # Transform output (only with regression)
        if self.mode == "regression":
            y = (y - self.settings["standardize"]["output"]["mean"]) / self.settings[
                "standardize"
            ]["output"]["std"]

        return x, y

    def _inverse_standardize(self, y: pd.Series) -> pd.Series:
        """
        For predictions, transform them back to application scales.
        Parameters
        ----------
        y [pd.Series]: Standardized output

        Returns
        -------
        y [pd.Series]: Actual output
        """
        assert self.settings["standardize"], "Standardize settings not found"
        return (
            y * self.settings["standardize"]["output"]["std"]
            + self.settings["standardize"]["output"]["mean"]
        )

    @staticmethod
    def _sort_results(results: pd.DataFrame) -> pd.DataFrame:
        return results.sort_values("worst_case", ascending=False)

    def _get_best_params(self, model, feature_set: str) -> dict:
        # Filter results for model and version
        results = self.results[
            np.logical_and(
                self.results["model"] == type(model).__name__,
                self.results["version"] == self.version,
            )
        ]

        # Filter results for feature set & sort them
        results = self._sort_results(results[results["dataset"] == feature_set])

        # Warning for unoptimized results
        if "Hyper Parameter" not in results["type"].values:
            warn("Hyper parameters not optimized for this combination")

        # Parse & return best parameters (regardless of if it's optimized)
        return utils.io.parse_json(results.iloc[0]["params"])

    def _grid_search_iteration(
        self, model, parameter_set: Union[str, None], feature_set: str
    ):
        """
        INTERNAL | Grid search for defined model, parameter set and feature set.
        """
        self.logger.info(
            f"Starting Hyper Parameter Optimization for {type(model).__name__} on "
            f"{feature_set} features ({len(self.x)} samples, "
            f"{len(self.feature_sets[feature_set])} features)"
        )

        # Select right hyper parameter optimizer
        if self.grid_search_type == "exhaustive":
            grid_search = ExhaustiveGridSearch(
                model,
                params=parameter_set,
                cv=self.cv,
                scoring=self.objective,
                n_trials=self.n_trials_per_grid_search,
                timeout=self.grid_search_timeout,
                verbose=self.verbose,
            )
        elif self.grid_search_type == "halving":
            grid_search = HalvingGridSearch(
                model,
                params=parameter_set,
                cv=self.cv,
                scoring=self.objective,
                n_trials=self.n_trials_per_grid_search,
                verbose=self.verbose,
            )
        elif self.grid_search_type == "optuna":
            grid_search = OptunaGridSearch(
                model,
                timeout=self.grid_search_timeout,
                cv=self.cv,
                n_trials=self.n_trials_per_grid_search,
                scoring=self.objective,
                verbose=self.verbose,
            )
        else:
            raise NotImplementedError(
                "Only Exhaustive, Halving and Optuna are implemented."
            )
        # Get results
        results = grid_search.fit(self.x[self.feature_sets[feature_set]], self.y)
        results = results.sort_values("worst_case", ascending=False)

        # Warn when best hyperparameters are close to predefined grid
        edge_params = grid_search.get_parameter_min_max()
        best_params = pd.Series(results["params"].iloc[0], name="best")
        params = edge_params.join(best_params, how="inner")

        def warn_when_too_close_to_edge(param: pd.Series, tol=0.01):
            # Min-max scaling
            min_, best, max_ = param["min"], param["best"], param["max"]
            scaled = (best - min_) / (max_ - min_)
            # Check if too close and warn if so
            if not (tol < scaled < 1 - tol):
                msg = "Optimal value for parameter is very close to edge case: "
                if min(abs(min_), abs(max_)) < 1:
                    msg += f"{param.name}={best:.2e} (range: {min_:.2e}...{max_:.2e})"
                else:
                    msg += f"{param.name}={best} (range: {min_}...{max_})"
                warn(msg)

        params.apply(warn_when_too_close_to_edge, axis=1)

        return results

    def _get_main_predictors(self, data):
        """
        Using Shapely Additive Explanations, this function calculates the main
        predictors for a given prediction and sets them into the class' memory.
        """
        # shap.TreeExplainer is not implemented for all models. So we try and fall back
        # to the feature importance given by the feature processor.
        # Note that the error would be raised when calling `TreeExplainer(best_model)`.
        try:
            # Get shap values
            best_model = self.best_model
            if type(best_model).__module__.startswith("amplo"):
                best_model = best_model.model
            # Note: The error would be raised at this point.
            #  So we have not much overhead.
            shap_values = np.array(TreeExplainer(best_model).shap_values(data))

            # Average over classes if necessary
            if shap_values.ndim == 3:
                shap_values = np.mean(np.abs(shap_values), axis=0)

            # Average over samples
            shap_values = np.mean(np.abs(shap_values), axis=0)
            shap_values /= shap_values.sum()  # normalize to sum up to 1
            idx_sort = np.flip(np.argsort(shap_values))

            # Set class attribute
            main_predictors = {
                col: score
                for col, score in zip(data.columns[idx_sort], shap_values[idx_sort])
            }

        except Exception:  # the exception can't be more specific  # noqa
            # Get shap feature importance
            fi = self.settings["feature_processing"]["feature_importance_"]["shap"]

            # Use only those columns that are present in the data
            main_predictors = {}
            missing_columns = []
            for col in data:
                if col in fi:
                    main_predictors[col] = fi[col]
                else:
                    missing_columns.append(col)

            if missing_columns:
                self.logger.warning(
                    f"Some data column names are missing in the shap feature "
                    f"importance dictionary: {missing_columns}"
                )

        # Some feature names are obscure since they come from the feature processing
        # module. Here, we relate the feature importance back to the original features.
        translation = translate_features(list(main_predictors))
        scores = {}
        for key, features in translation.items():
            for feat in features:
                scores[feat] = scores.get(feat, 0.0) + main_predictors[key]
        # Normalize
        total_score = np.sum(list(scores.values()))
        for key in scores:
            scores[key] /= total_score

        # Set attribute
        self._main_predictors = scores

        # Add to settings: [{"feature": "feature_name", "score": 1}, ...]
        scores_df = pd.DataFrame({"feature": scores.keys(), "score": scores.values()})
        scores_df.sort_values("score", ascending=False, inplace=True)
        self.settings["main_predictors"] = scores_df.to_dict("records")

        return scores



#  Copyright (c) 2022 by Amplo.

"""
Enables connection to Databricks via API calls.
"""


from __future__ import annotations

import os

from amplo.api._base import BaseRequestAPI

__all__ = ["DatabricksJobsAPI"]


_DATABRICKS_HOST_OS = "DATABRICKS_HOST"
_DATABRICKS_TOKEN_OS = "DATABRICKS_TOKEN"


class DatabricksJobsAPI(BaseRequestAPI):
    """
    Helper class for working with Databricks' Job API.

    References
    ----------
    [API reference](https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/latest/jobs)

    Parameters
    ----------
    host : str
        Databricks host (see https://docs.microsoft.com/en-us/azure/databricks/dev-tools/api/latest/authentication).
    access_token : str
        Access token (see Databricks > User Settings > Access tokens).
    """

    def __init__(self, host: str, access_token: str):
        super().__init__(host, access_token)

    def _authorization_header(self) -> dict:
        return {"Authorization": f"Bearer {self.access_token}"}

    @classmethod
    def from_os_env(
        cls, host_os: str | None = None, access_token_os: str | None = None
    ) -> DatabricksJobsAPI:
        """
        Instantiate the class using os environment strings.

        Parameters
        ----------
        host_os : str, default: _DATABRICKS_HOST_OS
            Key in the os environment for the Databricks host.
        access_token_os : str, default: _DATABRICKS_TOKEN_OS
            Key in the os environment for the Databricks access token.

        Returns
        -------
        DatabricksJobsAPI

        Raises
        ------
        KeyError
            When a os variable is not set.
        """

        host_os = host_os or _DATABRICKS_HOST_OS
        access_token_os = access_token_os or _DATABRICKS_TOKEN_OS

        host = os.environ[host_os]
        access_token = os.environ[access_token_os]
        return cls(host, access_token)

    def request(self, method: str, action: str, body: dict | None = None) -> dict:
        """
        Send a request to Databricks.

        Notes
        -----
        Every Databricks Jobs API call can be requested from here. In case you need more
        functionality than already implemented, have a look at the [OpenAPI
        specification](https://docs.microsoft.com/en-us/azure/databricks/_static/api-refs/jobs-2.1-azure.yaml).

        Parameters
        ----------
        method : str
            Request method (``GET``, ``PUT``, ``POST``, ...).
        action : str
            Path to Databricks Job (see Databricks Jobs API specification).
        body : dict of {str: int or str}, optional
            Request body schema (see Databricks Jobs API specification).

        Returns
        -------
        dict or {str: int or str}
            The request's response.

        Raises
        ------
        requests.HTTPError
            When the request's response has another status code than 200.
        """

        return super().request(method, action, json=body).json()

    # --------------------------------------------------------------------------
    # Jobs API requests

    def list_jobs(
        self,
        limit: int | None = None,
        offset: int | None = None,
        expand_tasks: bool | None = None,
    ) -> dict:
        body = {"limit": limit, "offset": offset, "expand_tasks": expand_tasks}
        return self.request("get", "2.1/jobs/list", body)

    def list_runs(
        self,
        limit: int | None = None,
        offset: int | None = None,
        expand_tasks: bool | None = None,
        *,
        active_only: bool | None = None,
        completed_only: bool | None = None,
        job_id: int | None = None,
        start_time_from: int | None = None,
        start_time_to: int | None = None,
    ) -> dict:
        body = {
            "limit": limit,
            "offset": offset,
            "expand_tasks": expand_tasks,
            "active_only": active_only,
            "completed_only": completed_only,
            "job_id": job_id,
            "start_time_from": start_time_from,
            "start_time_to": start_time_to,
        }
        return self.request("get", "2.1/jobs/runs/list", body)

    def get_job(self, job_id: int) -> dict:
        return self.request("get", "2.1/jobs/get", {"job_id": job_id})

    def get_run(self, run_id: int, include_history: bool | None = None) -> dict:
        body = {"run_id": run_id, include_history: include_history}
        return self.request("get", "2.1/jobs/runs/get", body)

    def run_job(
        self,
        job_id: int,
        idempotency_token: str | None = None,
        notebook_params: dict[str, int | str] | None = None,
        # Note: more parameters are available
    ) -> dict:
        body = {
            "job_id": job_id,
            "idempotency_token": idempotency_token,
            "notebook_params": notebook_params,
        }
        return self.request("post", "2.1/jobs/run-now", body)

    def cancel_run(self, run_id: int) -> dict:
        return self.request("post", "2.1/jobs/runs/cancel", {"run_id": run_id})



#  Copyright (c) 2022 by Amplo.

import re
import warnings

import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from sklearn.preprocessing import LabelEncoder

from amplo.utils import clean_feature_name
from amplo.utils.data import clean_keys
from amplo.utils.logging import get_root_logger

__all__ = ["DataProcessor"]


logger = get_root_logger().getChild("DataProcessor")


class DataProcessor:
    def __init__(
        self,
        target: str = None,
        float_cols: list = None,
        int_cols: list = None,
        date_cols: list = None,
        cat_cols: list = None,
        include_output: bool = True,
        missing_values: str = "interpolate",
        outlier_removal: str = "clip",
        z_score_threshold: int = 4,
        remove_constants: bool = True,
        version: int = 1,
        verbosity: int = 1,
    ):
        """
        Preprocessing Class. Cleans a dataset into a workable format.
        Deals with Outliers, Missing Values, duplicate rows, data types (floats,
        categorical and dates), Not a Numbers, Infinities.

        Parameters
        ----------
        target : str
            Column name of target variable
        float_cols : list
            Float columns
        int_cols : list
            Integer columns
        date_cols : list
            Date columns, all parsed to pd.datetime format
        cat_cols : list
            Categorical Columns. Currently, all one-hot encoded.
        include_output : bool
            Whether to include output in the data
        missing_values : str
            How to deal with missing values ("remove", "interpolate"or "mean")
        outlier_removal : str
            How to deal with outliers ("clip", "quantiles", "z-score"or "none")
        z_score_threshold : int
            If outlierRemoval="z-score", the threshold is adaptable
        remove_constants : bool
            If False, does not remove constant columns
        version : int
            Versioning the output files
        verbosity : int
            How much to print
        """
        # Tests
        mis_values_algo = ["remove_rows", "remove_cols", "interpolate", "mean", "zero"]
        if missing_values not in mis_values_algo:
            raise ValueError(
                "Missing values algorithm not implemented, pick from"
                f" {mis_values_algo}"
            )
        out_rem_algo = ["quantiles", "z-score", "clip", "none"]
        if outlier_removal not in out_rem_algo:
            raise ValueError(
                "Outlier Removal algorithm not implemented, pick from"
                f" {out_rem_algo}"
            )

        # Arguments
        self.version = version
        self.includeOutput = include_output
        self.target = (
            target if target is None else re.sub("[^a-z0-9]", "_", target.lower())
        )
        self.float_cols = (
            []
            if float_cols is None
            else [re.sub("[^a-z0-9]", "_", fc.lower()) for fc in float_cols]
        )
        self.int_cols = (
            []
            if int_cols is None
            else [re.sub("[^a-z0-9]", "_", ic.lower()) for ic in int_cols]
        )
        self.num_cols = self.float_cols + self.int_cols
        self.cat_cols = (
            []
            if cat_cols is None
            else [re.sub("[^a-z0-9]", "_", cc.lower()) for cc in cat_cols]
        )
        self.date_cols = (
            []
            if date_cols is None
            else [re.sub("[^a-z0-9]", "_", dc.lower()) for dc in date_cols]
        )
        if self.target in self.num_cols:
            self.num_cols.remove(self.target)

        # Algorithms
        self.missing_values = missing_values
        self.outlier_removal = outlier_removal
        self.z_score_threshold = z_score_threshold
        self.removeConstants = remove_constants

        # Fitted Settings
        self.data = None
        self.dummies = {}
        self._q1 = None
        self._q3 = None
        self._means = None
        self._stds = None
        self._label_encodings = []

        # Info for Documenting
        self.is_fitted = False
        self.verbosity = verbosity
        self.removedDuplicateRows = 0
        self.removedDuplicateColumns = 0
        self.removedOutliers = 0
        self.imputedMissingValues = 0
        self.removedConstantColumns = 0

    def _fit_transform(
        self, data: pd.DataFrame, fit=False, remove_constants=True
    ) -> "DataProcessor":
        """
        Wraps behavior of both, fitting and transforming the DataProcessor.
        The function basically reduces duplicated code fragments of `self.fit_transform`
         and `self.transform`.

        Parameters
        ----------
        data : pd.DataFrame
            Input data
        fit : bool
            If True, it will fit the transformer, too
        remove_constants : bool
            If True, it will remove constants when fit

        Returns
        -------
        DataProcessor
        """

        # Clean Keys
        self.data = clean_keys(data)

        # Impute columns
        self._impute_columns()

        # Remove target
        if (
            fit
            and not self.includeOutput
            and self.target is not None
            and self.target in self.data
        ):
            self.data = self.data.drop(self.target, axis=1)

        if fit:
            # Remove Duplicates
            self.remove_duplicates()

            # Infer data-types
            self.infer_data_types()

        # Convert data types
        self.convert_data_types(fit_categorical=fit)

        # Remove outliers
        self.remove_outliers(fit=fit)

        # Remove missing values
        self.remove_missing_values()

        # Remove Constants
        if fit and remove_constants:
            self.remove_constants()

        # Convert integer columns
        self.convert_float_int()

        # Encode or decode target
        self._code_target_column(fit=fit)

        return self

    def fit_transform(self, data: pd.DataFrame, remove_constants=True) -> pd.DataFrame:
        """
        Fits this data cleaning module and returns the transformed data.

        Parameters
        ----------
        data : pd.DataFrame
            Input data
        remove_constants : bool
            If True, it will remove constants when fit

        Returns
        -------
        pd.DataFrame
            Cleaned input data
        """
        if self.verbosity > 0:
            logger.info(
                f"Data Cleaning Started, ({len(data)} x {len(data.keys())}) samples"
            )

        self._fit_transform(data, fit=True, remove_constants=remove_constants)

        # Finish
        self.is_fitted = True
        if self.verbosity > 0:
            logger.info(
                f"Processing completed, ({len(self.data)} x {len(self.data.keys())})"
                " samples returned"
            )

        return self.data

    def transform(self, data: pd.DataFrame) -> pd.DataFrame:
        """
        Function that takes existing settings (including dummies), and transforms new
        data.

        Parameters
        ----------
        data : pd.DataFrame
            Input data

        Returns
        -------
        pd.DataFrame
            Cleaned input data
        """
        if not self.is_fitted:
            raise ValueError("Transform only available for fitted objects.")

        self._fit_transform(data, fit=False)

        return self.data

    def get_settings(self) -> dict:
        """
        Get settings to recreate fitted object.
        """
        assert self.is_fitted, "Object not yet fitted."
        settings = {
            "num_cols": self.num_cols,
            "float_cols": self.float_cols,
            "int_cols": self.int_cols,
            "date_cols": self.date_cols,
            "cat_cols": self.cat_cols,
            "_label_encodings": self._label_encodings,
            "missing_values": self.missing_values,
            "outlier_removal": self.outlier_removal,
            "z_score_threshold": self.z_score_threshold,
            "_means": None if self._means is None else self._means.to_json(),
            "_stds": None if self._stds is None else self._stds.to_json(),
            "_q1": None if self._q1 is None else self._q1.to_json(),
            "_q3": None if self._q3 is None else self._q3.to_json(),
            "dummies": self.dummies,
            "fit": {
                "imputed_missing_values": self.imputedMissingValues,
                "removed_outliers": self.removedOutliers,
                "removed_constant_columns": self.removedConstantColumns,
                "removed_duplicate_rows": self.removedDuplicateRows,
                "removed_duplicate_columns": self.removedDuplicateColumns,
            },
        }
        return settings

    def load_settings(self, settings: dict) -> None:
        """
        Loads settings from dictionary and recreates a fitted object
        """
        self.num_cols = settings.get("num_cols", [])
        self.float_cols = settings.get("float_cols", [])
        self.int_cols = settings.get("int_cols", [])
        self.date_cols = settings.get("date_cols", [])
        self.cat_cols = settings.get("cat_cols", [])
        self._label_encodings = settings.get("_label_encodings", [])
        self.missing_values = settings.get("missing_values", [])
        self.outlier_removal = settings.get("outlier_removal", [])
        self.z_score_threshold = settings.get("z_score_threshold", [])
        self._means = (
            None
            if settings["_means"] is None
            else pd.read_json(settings["_means"], typ="series")
        )
        self._stds = (
            None
            if settings["_stds"] is None
            else pd.read_json(settings["_stds"], typ="series")
        )
        self._q1 = (
            None
            if settings["_q1"] is None
            else pd.read_json(settings["_q1"], typ="series")
        )
        self._q3 = (
            None
            if settings["_q3"] is None
            else pd.read_json(settings["_q3"], typ="series")
        )
        self.dummies = settings.get("dummies", {})
        self.is_fitted = True

    def infer_data_types(self, data=None):
        """
        In case no data types are provided, this function infers the most likely data
        types
        """
        if len(self.cat_cols) == len(self.num_cols) == len(self.date_cols) == 0:
            # First cleanup
            self.data = (
                self.data.infer_objects() if data is None else data.infer_objects()
            )

            # Iterate through keys
            for key in self.data.keys():
                # Skip target
                if key == self.target:
                    continue

                # Integer
                elif pd.api.types.is_integer_dtype(self.data[key]):
                    self.int_cols.append(key)

                    if self.verbosity > 1:
                        logger.info(f"Found integer dtype: {key}")
                    continue

                # Float
                elif pd.api.types.is_float_dtype(self.data[key]):
                    self.float_cols.append(key)

                    if self.verbosity > 1:
                        logger.info(f"Found float dtype: {key}")
                    continue

                # Datetime
                elif pd.api.types.is_datetime64_any_dtype(self.data[key]):
                    self.date_cols.append(key)

                    if self.verbosity > 1:
                        logger.info(f"Found datetime dtype: {key}")
                    continue

                # Booleans
                elif pd.api.types.is_bool_dtype(self.data[key]):
                    self.int_cols.append(key)

                    if self.verbosity > 1:
                        logger.info(f"Found boolean dtype: {key}")
                    continue

                # Strings / Objects
                elif pd.api.types.is_object_dtype(self.data[key]):

                    # Check numerical
                    numeric = pd.to_numeric(
                        self.data[key], errors="coerce", downcast="integer"
                    )
                    if numeric.isna().sum() < len(self.data) * 0.3:
                        # Float
                        if pd.api.types.is_float_dtype(numeric):
                            self.float_cols.append(key)

                            if self.verbosity > 1:
                                logger.info(f"Found float dtype: {key}")

                        # Integer
                        if pd.api.types.is_integer_dtype(numeric):
                            self.int_cols.append(key)

                            if self.verbosity > 1:
                                logger.info(f"Found integer dtype: {key}")

                        # Update data and continue
                        self.data[key] = numeric
                        continue

                    # Check date (random subsample as it's expensive)
                    date = pd.to_datetime(
                        self.data[key].astype("str"),
                        errors="coerce",
                        infer_datetime_format=True,
                    )
                    if date.isna().sum() < 0.3 * len(self.data):
                        self.date_cols.append(key)
                        self.data[key] = date

                        if self.verbosity > 1:
                            logger.info(f"Found datetime dtype: {key}")
                        continue

                    # Check categorical variable
                    if self.data[key].nunique() < max(10, len(self.data) // 4):
                        self.cat_cols.append(key)

                        if self.verbosity > 1:
                            logger.info(f"Found categorical dtype: {key}")
                        continue

                # Else not found
                warnings.warn(f"Couldn't identify feature: {key}")

            # Set num cols for reverse compatibility
            self.num_cols = self.int_cols + self.float_cols

            # Print
            if self.verbosity > 0:
                logger.info(
                    f"Found {len(self.int_cols)} integer, {len(self.float_cols)} float,"
                    f" {len(self.cat_cols)} "
                    f"categorical and {len(self.date_cols)} datetime columns"
                )

        return

    def convert_data_types(
        self, data: pd.DataFrame = None, fit_categorical: bool = True
    ) -> pd.DataFrame:
        """
        Cleans up the data types of all columns.

        Parameters
        ----------
        data : pd.DataFrame
            Input data
        fit_categorical : bool
            Whether to fit the categorical encoder

        Returns
        -------
        pd.DataFrame
            Cleaned input data
        """
        # Set data
        if data is not None:
            self.data = data

        # Datetime columns
        for key in self.date_cols:
            self.data.loc[:, key] = pd.to_datetime(
                self.data[key], errors="coerce", infer_datetime_format=True, utc=True
            )

        # Integer columns
        for key in self.int_cols:
            self.data.loc[:, key] = pd.to_numeric(
                self.data[key], errors="coerce", downcast="integer"
            )

        # Float columns
        for key in self.float_cols:
            self.data.loc[:, key] = pd.to_numeric(
                self.data[key], errors="coerce", downcast="float"
            )

        # Categorical columns
        if fit_categorical:
            self.data = self._fit_cat_cols()
        else:
            assert self.is_fitted, (
                ".convert_data_types() was called with fit_categorical=False, while "
                "categorical encoder is not yet fitted."
            )
            self.data = self._transform_cat_cols()

        # We need everything to become numeric, so all that is not mentioned will be
        # handled as numeric
        all_cols = (
            self.float_cols
            + self.int_cols
            + self.date_cols
            + self.cat_cols
            + [self.target]
        )
        for key in self.data.keys():
            if key not in all_cols:
                self.data.loc[:, key] = pd.to_numeric(self.data[key], errors="coerce")

        return self.data

    def _fit_cat_cols(self, data: pd.DataFrame = None) -> pd.DataFrame:
        """
        Encoding categorical variables always needs a scheme. This fits the scheme.
        """
        if data is not None:
            self.data = data

        for key in self.cat_cols:
            # Clean the categorical variables
            self.data[key] = self.data[key].apply(clean_feature_name)

            # Get dummies, convert & store
            dummies = pd.get_dummies(
                self.data[key], prefix=key, dummy_na=self.data[key].isna().sum() > 0
            )
            self.data = pd.concat([self.data.drop(key, axis=1), dummies], axis=1)
            self.dummies[key] = dummies.keys().tolist()
        return self.data

    def _transform_cat_cols(self, data: pd.DataFrame = None) -> pd.DataFrame:
        """
        Converts categorical variables according to fitted scheme.
        """
        if data is not None:
            self.data = data

        for key in self.cat_cols:
            # Clean column
            self.data[key] = self.data[key].apply(clean_feature_name)

            # Transform
            value = self.dummies[key]
            dummies = [i[len(key) + 1 :] for i in value]
            self.data[value] = (
                np.equal.outer(self.data[key].astype("str").values, dummies) * 1
            )
            self.data = self.data.drop(key, axis=1)
        return self.data

    def remove_duplicates(
        self, data: pd.DataFrame = None, rows: bool = True
    ) -> pd.DataFrame:
        """
        Removes duplicate columns and rows.

        Parameters
        ----------
        rows : bool
            Whether to remove duplicate rows --> not desirable to maintain timelines
        data : pd.DataFrame
            Input data

        Returns
        -------
        pd.DataFrame
            Cleaned input data
        """
        if data is not None:
            self.data = data

        # Note down
        n_rows, n_columns = len(self.data), len(self.data.keys())

        # Remove Duplicates
        if rows:
            self.data = self.data.drop_duplicates()
        self.data = self.data.loc[:, ~self.data.columns.duplicated()]

        # Note
        self.removedDuplicateColumns = n_columns - len(self.data.keys())
        self.removedDuplicateRows = n_rows - len(self.data)
        if self.verbosity > 0 or (
            self.removedDuplicateColumns != 0 or self.removedDuplicateRows != 0
        ):
            logger.info(
                f"Removed {self.removedDuplicateColumns} duplicate columns and"
                f" {self.removedDuplicateRows} duplicate rows"
            )

        return self.data

    def remove_constants(self, data: pd.DataFrame = None) -> pd.DataFrame:
        """
        Removes constant columns
        """
        if data is not None:
            self.data = data
        columns = len(self.data.keys())

        # Remove Constants
        if self.removeConstants:
            const_cols = [
                col
                for col in self.data
                if self.data[col].nunique() == 1 and col != self.target
            ]
            self.data = self.data.drop(columns=const_cols)

        # Note
        self.removedConstantColumns = columns - len(self.data.keys())
        if self.verbosity > 0 or self.removedConstantColumns != 0:
            logger.info(f"Removed {self.removedConstantColumns} constant columns.")

        return self.data

    def fit_outliers(self, data: pd.DataFrame = None):
        """
        Checks outliers
        """
        if data is not None:
            self.data = data

        # With quantiles
        if self.outlier_removal == "quantiles":
            self._q1 = self.data[self.num_cols].quantile(0.25)
            self._q3 = self.data[self.num_cols].quantile(0.75)

        # By z-score
        elif self.outlier_removal == "z-score":
            self._means = self.data[self.num_cols].mean(skipna=True, numeric_only=True)
            self._stds = self.data[self.num_cols].std(skipna=True, numeric_only=True)
            self._stds[self._stds == 0] = 1

    def remove_outliers(
        self, data: pd.DataFrame = None, fit: bool = True
    ) -> pd.DataFrame:
        """
        Removes outliers
        """
        if data is not None:
            self.data = data

        # Check if needs fitting
        if fit:
            self.fit_outliers(self.data)
        else:
            if not self.is_fitted:
                raise ValueError(
                    ".remove_outliers() is called with fit=False, yet the object isn't"
                    " fitted yet."
                )

        # With Quantiles
        if self.outlier_removal == "quantiles":
            self.removedOutliers = (
                (self.data[self.num_cols] > self._q3).sum().sum()
                + (self.data[self.num_cols] < self._q1).sum().sum()
            ).tolist()
            self.data[self.num_cols] = self.data[self.num_cols].mask(
                self.data[self.num_cols] < self._q1
            )
            self.data[self.num_cols] = self.data[self.num_cols].mask(
                self.data[self.num_cols] > self._q3
            )

        # With z-score
        elif self.outlier_removal == "z-score":
            z_score = abs((self.data[self.num_cols] - self._means) / self._stds)
            self.removedOutliers = (
                (z_score > self.z_score_threshold).sum().sum().tolist()
            )
            self.data[self.num_cols] = self.data[self.num_cols].mask(
                z_score > self.z_score_threshold
            )

        # With clipping
        elif self.outlier_removal == "clip":
            self.removedOutliers = (
                (self.data[self.num_cols] > 1e12).sum().sum()
                + (self.data[self.num_cols] < -1e12).sum().sum()
            ).tolist()
            self.data[self.num_cols] = self.data[self.num_cols].clip(
                lower=-1e12, upper=1e12
            )
        return self.data

    def remove_missing_values(self, data: pd.DataFrame = None) -> pd.DataFrame:
        """
        Fills missing values (infinities and "not a number"s)
        """
        if data is not None:
            self.data = data

        # Replace infinities
        self.data = self.data.replace([np.inf, -np.inf], np.nan)

        # Note
        self.imputedMissingValues = self.data[self.num_cols].isna().sum().sum().tolist()
        if self.verbosity > 0 or self.imputedMissingValues != 0:
            logger.info(f"Imputed {self.imputedMissingValues} missing values.")

        # Removes all rows with missing values
        if self.missing_values == "remove_rows":
            self.data = self.data[self.data.isna().sum(axis=1) == 0]

        # Removes all columns with missing values
        elif self.missing_values == "remove_cols":
            self.data = self.data.loc[:, self.data.isna().sum(axis=0) == 0]

        # Fills all missing values with zero
        elif self.missing_values == "zero":
            self.data = self.data.fillna(0)

        # Linearly interpolates missing values
        elif self.missing_values == "interpolate":

            # Columns which are present with >10% missing values are not interpolated
            zero_keys = self.data.keys()[
                self.data.isna().sum() / len(self.data) > 0.1
            ].tolist()

            # Get all non-date_cols & interpolate
            ik = np.setdiff1d(self.data.keys().to_list(), self.date_cols + zero_keys)
            self.data[ik] = self.data[ik].interpolate(limit_direction="both")

            # Fill date columns
            for key in self.date_cols:
                if self.data[key].isna().sum() != 0:
                    # Interpolate
                    ints = pd.Series(self.data[key].values.astype("int64"))
                    ints[ints < 0] = np.nan
                    self.data[key] = pd.to_datetime(ints.interpolate(), unit="ns")

            # Fill rest (date & more missing values cols)
            if self.data.isna().sum().sum() != 0:
                self.data = self.data.fillna(0)

        # Fill missing values with column mean
        elif self.missing_values == "mean":
            self.data = self.data.fillna(self.data.mean())

            # Need to be individual for some reason
            for key in self.date_cols:
                self.data[key] = self.data[key].fillna(self.data[key].mean())

        return self.data

    def convert_float_int(self, data: pd.DataFrame = None) -> pd.DataFrame:
        """
        Integer columns with NaN in them are interpreted as floats.
        In the beginning we check whether some columns should be integers,
        but we rely on different algorithms to take care of the NaN.
        Therefore, we only convert floats to integers at the very end
        """
        if data is not None:
            self.data = data

        for key in self.int_cols:
            if key in self.data:
                self.data[key] = pd.to_numeric(
                    self.data[key], errors="coerce", downcast="integer"
                )
        return self.data

    def _code_target_column(self, fit):
        """En- or decodes target column of `self.data`

        Parameters
        ----------
        fit : bool
            Whether to fit the encoder
        """

        if self.target not in self.data:
            return

        # Get labels and encode / decode
        labels = self.data.loc[:, self.target]

        # Encode
        self.data.loc[:, self.target] = self.encode_labels(
            labels, fit=fit, warn_unencodable=False
        )

    def encode_labels(self, labels, *, fit=True, warn_unencodable=True):
        """Encode labels to numerical dtype

        Parameters
        ----------
        labels : np.ndarray or pd.Series
            Labels to encode
        fit : bool
            Whether to (re)fit the label encoder
        warn_unencodable : bool
            Whether to warn when labels are assumed being for regression task

        Returns
        -------
        np.ndarray
            Encoded labels or original labels if unencodable

        Raises
        ------
        NotFittedError
            When no label encoder has yet been trained, i.e. `self._label_encodings` is
            empty
        """
        # Convert to pd.Series for convenience
        labels = pd.Series(labels)

        # It"s probably a classification task
        if labels.dtype == object or labels.nunique() <= labels.size / 2:
            # Create encoding
            encoder = LabelEncoder()
            if fit is True:
                # Fit
                encoder.fit(labels)
                self._label_encodings = pd.Series(encoder.classes_).to_list()
            elif not self._label_encodings:
                raise NotFittedError("Encoder it not yet fitted")
            else:
                encoder.fit(self._label_encodings)
            # Encode
            return encoder.transform(labels)

        # It"s probably a regression task, thus no encoding needed
        if warn_unencodable:
            warnings.warn(
                UserWarning(
                    "Labels are probably for regression. No encoding happened..."
                )
            )
        return labels.to_numpy()

    def decode_labels(self, labels, *, except_not_fitted=True):
        """Decode labels from numerical dtype to original value

        Parameters
        ----------
        labels : np.ndarray or pd.Series
            Labels to decode
        except_not_fitted : bool
            Whether to raise an exception when label encoder is not fitted

        Returns
        -------
        np.ndarray
            Decoded labels or original labels if label encoder is not fitted and
            `except_not_fitted` is True

        Raises
        ------
        NotFittedError
            When `except_not_fitted` is True and label encoder is not fitted
        """
        try:
            if len(self._label_encodings) == 0:
                raise NotFittedError(
                    "Encoder it not yet fitted. Try first calling `encode_target` "
                    "to set an encoding"
                )
            encoder = LabelEncoder()
            encoder.classes_ = np.array(self._label_encodings)
            return encoder.inverse_transform(labels)
        except NotFittedError as err:
            if except_not_fitted:
                raise err
            else:
                return labels.to_numpy() if isinstance(labels, pd.Series) else labels

    def _impute_columns(self, data: pd.DataFrame = None) -> pd.DataFrame:
        """
        *** For production ***
        If a dataset is missing certain columns, this function looks at all registered
        columns and fills them with
        zeros.
        """
        if data is not None:
            self.data = data

        imputed = []
        for keys in [self.num_cols, self.date_cols, self.cat_cols]:
            for key in [k for k in keys if k not in self.data]:
                self.data[key] = np.zeros(len(self.data))
                imputed.append(key)
        if len(imputed) > 0:
            warnings.warn(f"Imputed {len(imputed)} missing columns! {imputed}")
        return self.data

    def prune_features(self, features: list):
        """
        For use with AutoML.Pipeline. We practically never use all features. Yet this
        processor imputes any missing features. This causes redundant operations,
        memory, and warnings. This function prunes the features to avoid that.

        parameters
        ----------
        features : list
            Required features (NOTE: include required features for extracted)
        """
        hash_features = dict([(k, 0) for k in features])
        self.date_cols = [f for f in self.date_cols if f in hash_features]
        self.num_cols = [f for f in self.num_cols if f in hash_features]
        self.int_cols = [f for f in self.int_cols if f in hash_features]
        self.float_cols = [f for f in self.float_cols if f in hash_features]
        self.cat_cols = [f for f in self.cat_cols if f in hash_features]



#  Copyright (c) 2022 by Amplo.

"""
Base class repository of Amplo.
"""

from amplo.base.exceptions import NotFittedError
from amplo.base.objects import (
    BaseEstimator,
    BaseObject,
    BasePredictor,
    BaseTransformer,
    LoggingMixin,
)



#  Copyright (c) 2022 by Amplo.

import warnings
from typing import Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy import stats

from amplo.utils import hist_search


class DataDriftWarning(Warning):
    pass


class DriftDetector:
    # todo add second order pdf fit
    # todo add subsequence drift detection

    def __init__(
        self,
        num_cols: list = None,
        cat_cols: list = None,
        date_cols: list = None,
        n_bins: int = 500,
        sigma: int = 3,
        with_pdf: bool = False,
    ):
        """
        Detects data drift in streamed input data.
        Supports numerical, categorical and datetime variables.
        Due to streamed, we don't check distributions, just bins.
        Categorical simply checks whether it's not a new column
        Datetime simply checks whether the date is recent
        """
        # Copy kwargs
        self.num_cols = [] if num_cols is None else num_cols
        self.cat_cols = [] if cat_cols is None else cat_cols
        self.date_cols = [] if date_cols is None else date_cols
        self.n_bins = n_bins
        self.with_pdf = with_pdf
        self.sigma = sigma

        # Initialize
        self.bins = {}
        for key in self.cat_cols:
            self.bins[key] = {}
        for key in self.num_cols:
            self.bins[key] = (None, None)
        self.output_bins = (None, None)
        self.distributions = {}

    def fit(self, data: pd.DataFrame) -> "DriftDetector":
        """
        Fits the class object
        """
        # Numerical
        self._fit_bins(data)
        self._fit_distributions(data)

        # Categorical

        return self

    def check(self, data: pd.DataFrame):
        """
        Checks a new dataframe for distribution drift.
        """
        violations = []

        violations.extend(self._check_bins(data))
        violations.extend(self._check_distributions(data))

        return violations

    def fit_output(self, model, data: pd.DataFrame):
        """
        Additionally to detecting input drift, we should also detect output drift. When
        the distribution of predicted outcomes change, it's often a sign that some under
        laying dynamics are shifting.
        """
        assert hasattr(model, "predict"), "Model does not have 'predict' attribute."

        # If it's a classifier and has predict_proba, we use that :)
        if hasattr(model, "predict_proba"):
            prediction = model.predict_proba(data)[:, 1]
        else:
            prediction = model.predict(data)

        ma, mi = max(prediction), min(prediction)
        y, x = np.histogram(
            prediction,
            bins=self.n_bins,
            range=(mi - (ma - mi) / 10, ma + (ma - mi) / 10),
        )
        self.output_bins = (x.tolist(), y.tolist())

    def check_output(self, model, data: pd.DataFrame, add: bool = False):
        """
        Checks the predictions of a model.
        """
        assert hasattr(model, "predict"), "Model does not have 'predict' attribute."

        # If it's a classifier and has predict_proba, we use that :)
        if hasattr(model, "predict_proba"):
            prediction = model.predict_proba(data)[:, 1]
        else:
            prediction = model.predict(data)

        # Check all predictions
        x, y = self.output_bins
        count_drifts = 0
        for value in prediction:
            ind = hist_search(x, value)
            if ind == -1 or y[ind] <= 0:
                # Drift detected
                count_drifts += 1
        if count_drifts > 0:
            severity = count_drifts / len(prediction) * 100
            warnings.warn(
                DataDriftWarning(f"Output drift detected! Severity: {severity:.2f}%")
            )

        # Add new output
        if add:
            y += np.histogram(prediction, bins=x)
            return y

    def get_weights(self) -> dict:
        """
        Gets the weights of the fitted object.
        Useful to save :)
        """
        return {
            "bins": self.bins,
            "output_bins": self.output_bins,
            "distributions": self.distributions,
        }

    def load_weights(self, weights):
        """
        Sets the weights of the object to recreate a previously fitted object.

        Parameters
        ----------
        weights : typing.Dict[str, dict or tuple]
            Weights of a (fitted) object.
            Expected keys are:
            - "bins" (dict):
                Bins dictionary with bins and quantities for all numeric keys.
            - "output_bins" (tuple):
                Output bins.
            - "distributions" (dict):
                Fitted distributions for all numeric keys.
        """
        self.bins = weights.get("bins", {})
        self.output_bins = weights.get("output_bins", (None, None))
        self.distributions = weights.get("distributions", {})
        return self

    def _fit_bins(self, data: pd.DataFrame):
        """
        Fits a histogram on each numerical column.
        """
        # Fit numerical
        for key in self.num_cols:
            ma, mi = data[key].max(), data[key].min()
            y, x = np.histogram(
                data[key],
                bins=self.n_bins,
                range=(mi - (ma - mi) / 10, ma + (ma - mi) / 10),
            )
            self.bins[key] = (x.tolist(), y.tolist())

        # Fit categorical
        for key in self.cat_cols:
            self.bins[key] = data[key].value_counts().to_dict()

    def _check_bins(self, data: pd.DataFrame, add: bool = False):
        """
        Checks if the current data falls into bins
        """
        violations = []

        for key in self.num_cols:
            # Get bins
            x, y = self.bins[key]

            # Check bins
            if isinstance(data, pd.DataFrame):
                for value in data[key].values:
                    ind = hist_search(x, value)
                    if ind == -1 or y[ind] <= 0:
                        violations.append(key)
                        break
            elif isinstance(data, pd.Series):
                ind = hist_search(x, data[key])
                if ind == -1 or (
                    y[ind] <= 0
                    and y[max(0, ind - 1)] <= 0
                    and y[min(self.n_bins, ind + 1)] <= 0
                ):
                    violations.append(key)

            # Add data
            if add:
                y += np.histogram(data[key], bins=x)
                self.bins[key] = (x, y)

        if len(violations) > 0:
            warnings.warn(
                DataDriftWarning(
                    f"Drift detected! "
                    f"{len(violations)} features outside training bins: {violations}"
                )
            )

        return violations

    def _fit_distributions(self, data: pd.DataFrame):
        """
        Fits a distribution on each numerical column.
        """
        if self.with_pdf:
            distributions = ["gamma", "beta", "dweibull", "dgamma"]
            distances = []
            fitted = []

            # Iterate through numerical columns
            for key in self.num_cols:
                y, x = np.histogram(data[key], normed=True)
                x = (x + np.roll(x, -1))[:-1] / 2.0  # Get bin means

                # Iterate through distributions
                for distribution in distributions:
                    # Fit & Get PDF
                    dist = getattr(stats, distribution)

                    # Multiple order fit
                    params = dist.fit(data[key])
                    fitted_pdf = dist.pdf(
                        x, loc=params[-2], scale=params[-1], *params[:-2]
                    )

                    # Analyse
                    distances.append(sum((y - fitted_pdf) ** 2))
                    fitted.append(
                        {
                            "distribution": distribution,
                            "params": params,
                        }
                    )
                plt.legend(["Original"] + distributions)
                plt.show()

                # Select lowest
                self.distributions[key] = fitted[np.argmin(distances)]

    def _check_distributions(self, data: pd.DataFrame) -> list:
        """
        Checks whether the new data falls within the fitted distributions
        """
        # Init
        violations = []

        if self.with_pdf:
            # Check all numerical columns
            for key in self.num_cols:
                dist = getattr(stats, self.distributions[key]["distribution"])
                params = self.distributions[key]["params"]
                probabilities = dist.pdf(
                    data[key].values, loc=params[-2], scale=params[-1], *params[:-2]
                )

                if any(p < self.sigma for p in probabilities):
                    violations.append(key)
                    continue

            if len(violations) > 0:
                warnings.warn(
                    f"Drift detected! {len(violations)} features outside training bins:"
                    f" {violations}",
                    DataDriftWarning,
                )

        return violations

    def add_output_bins(
        self, old_bins: tuple, prediction: Union[np.ndarray, pd.Series]
    ):
        """
        Just a utility, adds new data to an old distribution.
        """
        if len(old_bins) != 0:
            x, y = old_bins
            yn = np.histogram(prediction, bins=x)[0].tolist()
            y = [y[i] + yn[i] for i in range(len(y))]
        else:
            y, x = np.histogram(prediction, bins=self.n_bins)
            x, y = x.tolist(), y.tolist()
        return x, y

    def add_bins(self, bins: dict, data: pd.DataFrame):
        """
        Just a utility, adds new data to an old distribution.
        """
        # Add numerical bins
        for key in self.num_cols:
            if key in data:
                if key in bins:
                    x, y = bins[key]
                elif key in self.bins:
                    x, _ = self.bins[key]
                    y = [0 for i in range(len(x) - 1)]
                else:
                    raise ValueError(f"Drift Detector - Adding unfitted feature. {key}")
                yn = np.histogram(data[key], bins=x)[0].tolist()
                y = [y[i] + yn[i] for i in range(len(y))]
                bins[key] = (x, y)

        # Add categorical
        for key in self.cat_cols:
            if key in data:
                counts = data[key].value_counts().to_dict()
                bins[key] = {
                    k: counts.get(k, 0) + bins.get(key, {}).get(k, 0)
                    for k in counts.keys() | bins.get(key, {}).keys()
                }
        return bins



#  Copyright (c) 2022 by Amplo.

from __future__ import annotations

import json
import os
import re
from pathlib import Path
from typing import TYPE_CHECKING, Iterable, cast
from warnings import warn

import pandas as pd
from requests import HTTPError

from amplo.api.storage import AzureBlobDataAPI

if TYPE_CHECKING:
    from amplo.api.platform import AmploPlatformAPI

__all__ = [
    "boolean_input",
    "parse_json",
    "read_pandas",
    "get_file_metadata",
    "merge_folders",
    "merge_logs",
]


FILE_READERS = {
    ".csv": pd.read_csv,
    ".json": pd.read_json,
    ".xml": pd.read_xml,
    ".feather": pd.read_feather,
    ".parquet": pd.read_parquet,
    ".stata": pd.read_stata,
    ".pickle": pd.read_pickle,
}


def boolean_input(question: str) -> bool:
    x = input(question + " [y / n]")
    if x.lower() == "n" or x.lower() == "no":
        return False
    elif x.lower() == "y" or x.lower() == "yes":
        return True
    else:
        warn('Sorry, I did not understand. Please answer with "n" or "y"')
        return boolean_input(question)


def parse_json(json_string: str | dict) -> str | dict:
    if isinstance(json_string, dict):
        return json_string
    else:
        try:
            return json.loads(
                json_string.replace("'", '"')
                .replace("True", "true")
                .replace("False", "false")
                .replace("nan", "NaN")
                .replace("None", "null")
            )
        except json.decoder.JSONDecodeError:
            warn(f"Cannot validate, impassable JSON: {json_string}")
            return json_string


def read_pandas(path: str | Path) -> pd.DataFrame:
    """
    Wrapper for various read functions

    Returns
    -------
    pd.DataFrame
    """
    file_extension = Path(path).suffix
    if file_extension not in FILE_READERS:
        raise NotImplementedError(f"File format {file_extension} not supported.")
    else:
        reader = FILE_READERS[file_extension]
        return reader(path, low_memory=False)


def get_file_metadata(file_path: str | Path) -> dict[str, str | float]:
    """
    Get file metadata from given path.

    Parameters
    ----------
    file_path : str or Path
        File path.

    Returns
    -------
    dict of {str: str or float}
        File metadata.

    Raises
    ------
    FileNotFoundError
        When the path does not exist.
    IsADirectoryError
        When the path resolves a directory, not a file.
    """

    from amplo.utils import check_dtypes

    check_dtypes("file_path", file_path, (str, Path))

    file_path = Path(file_path)

    if not file_path.exists():
        raise FileNotFoundError(f"File does not exist: '{file_path}'")
    if not file_path.is_file():
        raise IsADirectoryError(f"Path is not a file: '{file_path}'")

    return {
        "file_name": str(file_path.name),
        "full_path": str(file_path.resolve()),
        # "creation_time": os.path.getctime(str(file_path)),
        "last_modified": os.path.getmtime(str(file_path)),
    }


def _read_files_in_folders(
    folders: Iterable[str | Path],
    blob_api: AzureBlobDataAPI | None = None,
) -> tuple[list[str], list[pd.DataFrame], list[dict[str, str | float]]]:
    """
    Use pandas to read all non-hidden and non-empty files into a DataFrame.

    Parameters
    ----------
    folders : iterable of (str or Path)
        Directory names.
    blob_api : AzureBlobDataAPI or None, optional
        If None, tries to read data from local folder, else from Azure, by default None

    Returns
    -------
    data : pd.DataFrame
        All files of the folders merged into one multi-indexed DataFrame.
    metadata : list of dict of {str : str or float}
        Metadata of data.

    Raises
    ------
    FileNotFoundError
        When any directory is empty, or has no supported file type.
    """

    # Map folders to pathlib.Path object
    folders = list(map(Path, folders))
    folders = cast(list[Path], folders)  # type hint

    # Initialize
    file_names, data, metadata = [], [], []

    for folder in sorted(folders):

        # List all files
        if blob_api:
            files = list(map(Path, blob_api.ls_files(folder)))
        else:
            files = [f for f in folder.iterdir() if f.is_file()]

        # Remove hidden files
        hidden_files = [f for f in files if re.match(r"^\..*", f.name)]
        files = list(set(files) - set(hidden_files))

        # Remove unsupported file types
        unsupported_files = [f for f in files if f.suffix not in FILE_READERS]
        files = list(set(files) - set(unsupported_files))

        # Remove empty files
        if blob_api:
            empty_files = [f for f in files if blob_api.get_size(f) == 0]
        else:
            empty_files = [f for f in files if f.stat().st_size == 0]
        files = list(set(files) - set(empty_files))

        # Sanity check
        if not files:
            raise FileNotFoundError(f"Directory is empty: '{folder}'")

        # Read files
        for file in sorted(files):

            if blob_api:
                datum = blob_api.read_pandas(file)
                metadatum = blob_api.get_metadata(file)
            else:
                datum = read_pandas(file)
                metadatum = get_file_metadata(file)

            if isinstance(datum, pd.Series):
                datum = datum.to_frame()

            file_names.append(str(file))
            data.append(datum)
            metadata.append(metadatum)

    return file_names, data, metadata


def _map_datalogs_to_file_names(
    file_names: list[str], platform_api: AmploPlatformAPI | None = None
) -> list[dict]:
    """
    Get datalogs for every filename.

    Parameters
    ----------
    file_names : list of str
        Files names to get datalogs from - if available.
    platform_api : AmploPlatformAPI or None, optional
        API to get datlogs from, by default None

    Returns
    -------
    list of dict
        Datalogs for every filename.
    """

    if not platform_api:
        return []

    # It is assumed that the 6th and 5th path position of the (first) filename contains
    # the team and machine name, respectively, if you count from right to left.
    # E.g., "Team/Machine/data/Category/Issue/log_file.csv"
    try:
        team, machine = file_names[0].split("/")[-6:-4]
    except IndexError:
        warn(f"Couldn't resolve Team and Machine name from filename: '{file_names[0]}'")
        return []

    # Get datalog for each filename
    datalogs = []
    for fname in file_names:
        try:
            datalog = platform_api.get_datalog(team, machine, fname)
        except HTTPError:
            # No matching datalog found. Do still append it to preserve the order.
            datalog = {}

        datalogs.append(datalog)

    return datalogs


def _mask_intervals(
    datalogs: list[dict],
    data: list[pd.DataFrame],
    metadata: list[dict[str, str | float]],
) -> tuple[list[pd.DataFrame], list[dict[str, str | float]]]:
    """
    Masks the data with the intervals given by the datalogs.

    Parameters
    ----------
    datalogs : list of dict
        Datalogs dictionary that should contain the keys 'selected' and 'datetime_col'.
    data : list of pd.DataFrame
        Data for splitting.
    metadata : list of dict of {str : str or float}
        Metadata of data. For each data split it will be duplicated.

    Returns
    -------
    data_out : list of pd.DataFrame
        Selected data.
    metadata_out : list of dict of {str : str or float}
        Same metadata but duplicated for every split.

    Raises
    ------
    ValueError
        When no valid match for the start or stop time of the data interval was found,
        i.e. when the time difference is more than 1 second.
    """

    # Initialize
    data_out = []
    metadata_out = []

    for datalog, datum, metadatum in zip(datalogs, data, metadata):
        # Get intervals and timestamp column from datalog
        intervals = datalog.get("selected", [])
        ts_col = datalog.get("datetime_col", "")

        # If no interval is to be selected, append the whole datum
        if not intervals:
            data_out.append(datum)
            metadata_out.append(metadatum)

        # Prevent a KeyError when ts_col column is not present
        if ts_col not in datum.columns:
            warn(f"Cannot select intervals as the column '{ts_col}' is not present.")
            data_out.append(datum)
            metadata_out.append(metadatum)

        # Else, select and append each interval of the datum
        else:
            # Extract unix timestamps
            ts = datum[ts_col]
            if not pd.api.types.is_datetime64_any_dtype(ts):
                ts = pd.to_datetime(ts)
            ts = ts.astype(int) / 10**9  # convert to unix format

            # Extract intervals
            for interval in intervals:
                # Find closest timestamps
                ts_first, ts_last = interval
                first = (ts - ts_first).abs().argmin()
                last = (ts - ts_last).abs().argmin()

                # Except when time difference is too large
                if abs(ts.iloc[first] - ts_first) > 1:
                    raise ValueError(f"Could not find a timestamp close to {ts_first}")
                if abs(ts.iloc[last] - ts_last) > 1:
                    raise ValueError(f"Could not find a timestamp close to {ts_last}")

                # Select interval
                data_out.append(datum.iloc[first : last + 1])
                metadata_out.append(metadatum)

    return data_out, metadata_out


def _make_multiindex(
    data: list[pd.DataFrame],
    metadata: list[dict[str, str | float]],
    target_col: str = "labels",
) -> tuple[pd.DataFrame, dict[int, dict[str, str | float]]]:
    """
    Merge list of dataframes into one multiindexed one.

    Parameters
    ----------
    data : list of pd.DataFrame
        Data to be merged.
    metadata : list of dict of {str : str or float}
        Metadata of data.
    target_col : str, optional
        Target column name. Values are depicted by the folder name, by default "labels"

    Returns
    -------
    data : pd.DataFrame
        All files of the folders merged into one multi-indexed DataFrame.
    metadata : dict of {int : dict of {str : str or float}}
        Metadata of merged data.
        The first index level values of the data correspond to the keys in the metadata.

    Raises
    ------
    ValueError
        When any file already has a column named after target_col.
    """

    # Initialize
    index_count = 0
    data_out = []
    metadata_out = {}

    for datum, metadatum in zip(data, metadata):

        # Get file and folder name
        file_name = metadatum["file_name"]
        folder_name = Path(str(metadatum["full_path"])).parent.name

        # Set label column
        if target_col in datum.columns:
            raise ValueError(
                f"The target name '{target_col}' already exists in the data columns"
                f" of file '{file_name}'."
            )
        datum[target_col] = folder_name

        # Set multiindex
        index = pd.MultiIndex.from_product(
            [[index_count], datum.index.values], names=["log", "index"]
        )
        datum.set_index(index, inplace=True)

        # Add data and metadata, and increment
        data_out.append(datum)
        metadata_out[index_count] = metadatum
        index_count += 1

    # Finish: concatenate data
    return pd.concat(data_out), metadata_out


def merge_folders(
    folders: Iterable[str | Path],
    target_col: str = "labels",
    blob_api: AzureBlobDataAPI | None = None,
    platform_api: AmploPlatformAPI | None = None,
) -> tuple[pd.DataFrame, dict[int, dict[str, str | float]]]:
    """
    Combine log files from given directories into a multiindexed DataFrame.

    Parameters
    ----------
    folders : iterable of (str or Path)
        Directory paths to read data from.
    target_col : str, optional
        Target column name. Values are depicted by the folder name, by default "labels"
    blob_api : AzureBlobDataAPI or None, optional
        Azure api when not reading from local, by default None
    platform_api : AmploPlatformAPI or None, optional
        Platform api for selecting intervals, by default None

    Returns
    -------
    data : pd.DataFrame
        All files of the folders merged into one multi-indexed DataFrame.
    metadata : dict of {int : dict of {str : str or float}}
        Metadata of merged data.
        The first index level values of the data correspond to the keys in the metadata.
    """

    # Pandas read files
    fnames, data, metadata = _read_files_in_folders(folders, blob_api)

    # Select intervals when datalogs are available
    datalogs = _map_datalogs_to_file_names(fnames, platform_api)
    if datalogs:
        data, metadata = _mask_intervals(datalogs, data, metadata)

    # Concatenate data and make it multiindexed
    data, metadata = _make_multiindex(data, metadata, target_col)

    return data, metadata


def merge_logs(
    parent_folder: str | Path,
    target_col: str = "labels",
    *,
    more_folders: list[str | Path] | None = None,
    azure: tuple[str, str] | bool = False,
    platform: tuple[str, str] | bool | None = None,
) -> tuple[pd.DataFrame, dict[int, dict[str, str | float]]]:
    """
    Combine log files of all subdirectories into a multi-indexed DataFrame.

    The function can handle logs from a local directory as well as data coming from an
    Azure blob storage. For the latter case it is furthermore capable to select
    intervals using Amplo's datalogs.

    Notes
    -----
    Make sure that each protocol is located in a subdirectory whose name represents the
    respective label.

    An exemplary directory structure of ``parent_folder``:
        ``
        parent_folder
        ├─ Label_1
        │   ├─ Log_1.*
        │   └─ Log_2.*
        ├─ Label_2
        │   └─ Log_3.*
        └─ ...
        ``

    Parameters
    ----------
    parent_folder : str or Path
        Directory that contains subdirectories with tabular data files.
    target_col : str
        Target column name. Values are depicted by the folder name.
    more_folders : list of str or Path, optional
        Additional folder names with tabular data files to append.
    azure : (str, str) or bool, default: False
        Use this parameter to indicate that data is in Azure blob storage.
        If False, it is assumed that data origins from local directory.
        If True, the AzureBlobDataAPI is initialized with default OS env variables.
        Otherwise, it will use the tuple to initialize the api.
    platform : (str, str) or bool or None, default: None
        Use this parameter for selecting data according to Amplo's datalogs.
        If None, its value is set to bool(azure).
        If False, no datalogs information will be used.
        If True, the AmploPlatformAPI is initialized with default OS env variables.
        Otherwise, it will use the tuple to initialize the api.

    Returns
    -------
    data : pd.DataFrame
        All files of the folders merged into one multi-indexed DataFrame.
        Multi-index names are 'log' and 'index'.
    metadata : dict of {int : dict of {str : str or float}}
        Metadata of merged data.
    """

    from amplo.api.platform import AmploPlatformAPI
    from amplo.utils import check_dtypes

    check_dtypes("parent_folder", parent_folder, (str, Path))
    check_dtypes("target_col", target_col, str)
    check_dtypes("more_folders", more_folders, (type(None), list))

    # Get azure blob client
    if not azure:
        blob_api = None
    else:
        azure = azure if not isinstance(azure, bool) else tuple()
        blob_api = AzureBlobDataAPI.from_os_env(*azure)

    # Mirror azure parameter when platform is not set
    if platform is None:
        platform = bool(azure)
    # Get amplo platform client
    if not platform:
        platform_api = None
    else:
        platform = platform if not isinstance(platform, bool) else tuple()
        platform_api = AmploPlatformAPI.from_os_env(*platform)

    # Get child folders
    if not blob_api:
        folders = [folder for folder in Path(parent_folder).iterdir()]
    else:
        folders = blob_api.ls_folders(parent_folder)

    # Add more_folders
    if more_folders:
        folders += more_folders

    return merge_folders(folders, target_col, blob_api, platform_api)
